from pubtools.sign.operations import (
    ClearSignOperation,
    ContainerSignOperation,
    BlobSignOperation,
)


def test_containersign_operation_doc_argument():
    assert ContainerSignOperation.doc_arguments() == {
        "options": {
            "digests": {"description": "List of digest to sign"},
            "references": {"description": "List of references to sign"},
            "requester": {"description": "Requester of the signing operation"},
            "signing_keys": {
                "description": "Signing key short ids which should be used for signing",
                "type": "list",
                "required": "true",
            },
            "signing_key_names": {
                "description": "Signing key names which should be used for signing"
            },
            "task_id": {
                "description": "Usually pub task id, serves as identifier for in signing request"
            },
            "identity_references": {"description": "List of references to sign"},
        },
        "examples": {
            "digests": "",
            "references": "",
            "signing_keys": ["123"],
            "signing_key_names": ["key1"],
            "task_id": "",
            "requester": "",
            "identity_references": "",
        },
    }


def test_clearsign_operation_doc_argument():
    assert ClearSignOperation.doc_arguments() == {
        "options": {
            "inputs": {"type": "list", "description": "Signing data", "required": "true"},
            "signing_keys": {
                "type": "list",
                "description": "Signing key short ids which should be used for signing",
                "required": "true",
            },
            "signing_key_names": {
                "description": "Signing key names which should be used for signing"
            },
            "task_id": {
                "type": "str",
                "description": "Usually pub task id, serves as identifier for in signing request",
                "required": "true",
            },
            "repo": {
                "type": "str",
                "description": "Repository name",
                "required": "true",
            },
            "requester": {"description": "Requester of the signing operation"},
        },
        "examples": {
            "inputs": ["input1", "input2"],
            "signing_keys": ["123"],
            "signing_key_names": ["key1"],
            "task_id": "1",
            "requester": "",
            "repo": "repo",
        },
    }


def test_container_sign_to_dict():
    assert ContainerSignOperation(
        digests=["digest"],
        references=["references"],
        signing_keys=["sig-key"],
        signing_key_names=["sig-key-name"],
        task_id="task-id",
        requester="requester",
    ).to_dict() == dict(
        digests=["digest"],
        references=["references"],
        signing_keys=["sig-key"],
        signing_key_names=["sig-key-name"],
        task_id="task-id",
        requester="requester",
    )


def test_clear_sign_to_dict():
    assert ClearSignOperation(
        inputs=["input1"],
        signing_keys=["sig-key"],
        task_id="task-id",
        repo="repo",
        requester="requester",
    ).to_dict() == dict(
        inputs=["input1"],
        signing_keys=["sig-key"],
        task_id="task-id",
        repo="repo",
        requester="requester",
    )


def test_blob_sign_to_dict():
    assert BlobSignOperation(
        blobs=["blob"],
        signing_keys=["sig-key"],
        task_id="task-id",
        requester="requester",
        signing_key_names=["sig-key-name"],
    ).to_dict() == dict(
        blobs=["blob"],
        signing_keys=["sig-key"],
        task_id="task-id",
        requester="requester",
        signing_key_names=["sig-key-name"],
    )



from typing import Any, List, Optional

from ..models.msg import MsgError

from proton.handlers import MessagingHandler
import proton

IGNORED_ERRORS = [("amqp:connection:framing-error", "SSL Failure: Unknown error")]

LOG_HDR_EVT_FMT = "[EVNT: {}]"


class _MsgClient(MessagingHandler):
    def __init__(self, errors: List[MsgError]) -> None:
        super().__init__()
        self.errors = errors

    def _format_log_msg(self, msg: str, event: Optional[proton.Event] = None) -> str:
        if event:
            hdr = LOG_HDR_EVT_FMT.format(event.type)
        else:
            hdr = ""
        return hdr + " " + msg

    def on_error(self, event: proton.Event, source: Any = None) -> bool:
        condition = getattr(source, "condition", None) or getattr(source, "remote_condition", None)
        if not condition:
            return False
        if (condition.name, condition.description) in IGNORED_ERRORS:
            return False
        self.errors.append(
            MsgError(
                name=condition.name,
                description=condition.description,
                source=source,
            )
        )
        event.container.stop()
        return True

    def on_link_error(self, event: proton.Event) -> None:
        self.on_error(event, event.link)

    def on_session_error(self, event: proton.Event) -> None:
        self.on_error(event, event.session)

    def on_connection_error(self, event: proton.Event) -> None:
        self.on_error(event, event.connection)

    def on_transport_error(self, event: proton.Event) -> None:
        self.on_error(event, event.transport)



from __future__ import annotations

from dataclasses import field, dataclass

from typing import List, ClassVar, Any

from ..results.operation_result import OperationResult

from .base import SignOperation


@dataclass
class ContainerSignOperation(SignOperation):
    """ContainersSignOperation model class."""

    ResultType: ClassVar[OperationResult]
    digests: List[str] = field(
        metadata={"description": "List of digest to sign"}, default_factory=list
    )
    references: List[str] = field(
        metadata={"description": "List of references to sign"}, default_factory=list
    )
    task_id: str = field(
        metadata={
            "description": "Usually pub task id, serves as identifier for in signing request"
        },
        default="",
    )
    identity_references: List[str] = field(
        metadata={"description": "List of references to sign"}, default_factory=list
    )
    requester: str = field(
        metadata={"description": "Requester of the signing operation"}, default=""
    )

    def to_dict(self) -> dict[str, Any]:
        """Return a dict representation of the object."""
        return dict(
            digests=self.digests,
            references=self.references,
            signing_key_names=self.signing_key_names,
            signing_keys=self.signing_keys,
            task_id=self.task_id,
            requester=self.requester,
        )



from pubtools.sign.signers.msgsigner import ContainerSignResult, ClearSignResult, BlobSignResult


def test_containeroperation_result_to_dict():
    assert ContainerSignResult(
        results=["test"], signing_keys=["signing_key"], failed=False
    ).to_dict() == {
        "results": ["test"],
        "signing_keys": ["signing_key"],
        "failed": False,
    }


def test_clearoperation_result_to_dict():
    assert ClearSignResult(outputs=["test"], signing_keys=["signing_key"]).to_dict() == {
        "outputs": ["test"],
        "signing_keys": ["signing_key"],
    }


def test_bloboperation_result_to_dict():
    assert BlobSignResult(
        results=["test"], signing_keys=["signing_key"], failed=False
    ).to_dict() == {"results": ["test"], "signing_keys": ["signing_key"], "failed": False}



import os

from click.testing import CliRunner
import pytest
from unittest.mock import patch, ANY


from pubtools.sign.models.msg import MsgError
from pubtools.sign.signers.msgsigner import (
    MsgBatchSigner,
    MsgSignerResults,
    ContainerSignOperation,
    ContainerSignResult,
    ClearSignOperation,
    ClearSignResult,
    SignRequestType,
    msg_container_sign_main,
)
from pubtools.sign.conf.conf import load_config
from pubtools.sign.models.msg import MsgMessage
from pubtools.sign.results.signing_results import SigningResults


def test_msg_batch_container_sign(f_msg_batch_signer, f_config_msg_batch_signer_ok):
    f_msg_batch_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "ok"
    }
    f_msg_batch_signer.return_value.sign.return_value.operation_result.results = []
    f_msg_batch_signer.return_value.sign.return_value.operation_result.signing_keys = []
    f_msg_batch_signer.return_value.sign.return_value.operation.to_dict.return_value = {}
    result = CliRunner().invoke(
        msg_container_sign_main,
        [
            "--signing-key",
            "test-signing-key",
            "--digest",
            "some-digest",
            "--reference",
            "some-reference",
            "--task-id",
            "1",
            "--config-file",
            f_config_msg_batch_signer_ok,
            "--signer-type",
            "batch",
        ],
    )
    print(result.stdout)
    assert result.exit_code == 0, result.output


def test_msg_batch_container_sign_unsupported(f_msg_batch_signer, f_config_msg_batch_signer_ok):
    f_msg_batch_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "ok"
    }
    f_msg_batch_signer.return_value.sign.return_value.operation_result.results = []
    f_msg_batch_signer.return_value.sign.return_value.operation_result.signing_key = ""
    f_msg_batch_signer.return_value.sign.return_value.operation.to_dict.return_value = {}
    result = CliRunner().invoke(
        msg_container_sign_main,
        [
            "--signing-key",
            "test-signing-key",
            "--digest",
            "some-digest",
            "--reference",
            "some-reference",
            "--task-id",
            "1",
            "--config-file",
            f_config_msg_batch_signer_ok,
            "--signer-type",
            "unsuported",
        ],
    )
    print(result.stdout)
    assert result.exit_code == 2, result.output


def test_msg_container_sign_keyname(f_msg_batch_signer, f_config_msg_batch_signer_ok):
    f_msg_batch_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "ok"
    }
    f_msg_batch_signer.return_value.sign.return_value.operation_result.results = []
    f_msg_batch_signer.return_value.sign.return_value.operation_result.signing_keys = []
    f_msg_batch_signer.return_value.sign.return_value.operation.to_dict.return_value = {}
    result = CliRunner().invoke(
        msg_container_sign_main,
        [
            "--signing-key",
            "test-signing-key",
            "--signing-key-name",
            "test-signing-key-name",
            "--digest",
            "some-digest",
            "--reference",
            "some-reference",
            "--task-id",
            "1",
            "--config-file",
            f_config_msg_batch_signer_ok,
            "--signer-type",
            "batch",
        ],
    )
    print(result.stdout)
    assert result.exit_code == 0, result.output


def test_msg_container_sign_requester(f_msg_batch_signer, f_config_msg_batch_signer_ok):
    f_msg_batch_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "ok"
    }
    f_msg_batch_signer.return_value.sign.return_value.operation_result.results = []
    f_msg_batch_signer.return_value.sign.return_value.operation_result.signing_keys = []
    f_msg_batch_signer.return_value.sign.return_value.operation.to_dict.return_value = {}
    result = CliRunner().invoke(
        msg_container_sign_main,
        [
            "--signing-key",
            "test-signing-key",
            "--digest",
            "some-digest",
            "--reference",
            "some-reference",
            "--task-id",
            "1",
            "--config-file",
            f_config_msg_batch_signer_ok,
            "--requester",
            "test-requester",
            "--signer-type",
            "batch",
        ],
    )
    print(result.stdout)
    assert result.exit_code == 0, result.output


def test_msg_container_sign_error(f_msg_batch_signer, f_config_msg_batch_signer_ok):
    f_msg_batch_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "error",
        "error_message": "simulated error",
    }
    f_msg_batch_signer.return_value.sign.return_value.operation_result.results = []
    f_msg_batch_signer.return_value.sign.return_value.operation_result.signing_key = ""
    f_msg_batch_signer.return_value.sign.return_value.operation.to_dict.return_value = {}
    result = CliRunner().invoke(
        msg_container_sign_main,
        [
            "--signing-key",
            "test-signing-key",
            "--digest",
            "some-digest",
            "--reference",
            "some-reference",
            "--task-id",
            "1",
            "--config-file",
            f_config_msg_batch_signer_ok,
            "--signer-type",
            "batch",
        ],
    )
    print(result.stdout)
    assert result.exit_code == 1, result.output


def test_msg_container_sign_raw(f_msg_batch_signer, f_config_msg_batch_signer_ok):
    f_msg_batch_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "ok"
    }
    f_msg_batch_signer.return_value.sign.return_value.operation_result.results = [
        ({"i": 456, "msg": {"errors": [], "signed_claim": "signed"}}, {})
    ]
    f_msg_batch_signer.return_value.sign.return_value.operation_result.signing_key = ""
    result = CliRunner().invoke(
        msg_container_sign_main,
        [
            "--signing-key",
            "test-signing-key",
            "--digest",
            "some-digest",
            "--reference",
            "some-reference",
            "--task-id",
            "1",
            "--config-file",
            f_config_msg_batch_signer_ok,
            "--signer-type",
            "batch",
            "--raw",
        ],
    )
    print(result.stdout)
    assert result.exit_code == 0, result.output
    assert result.output == "signed\n"


def test_msg_container_sign_raw_error(f_msg_batch_signer, f_config_msg_batch_signer_ok):
    f_msg_batch_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "ok",
        "error_message": "",
    }
    f_msg_batch_signer.return_value.sign.return_value.operation_result.results = [
        (
            {
                "i": 456,
                "msg": {"errors": ['no signing key "test-signing-key"'], "signed_claim": None},
            },
            {},
        )
    ]
    f_msg_batch_signer.return_value.sign.return_value.operation_result.signing_key = ""
    result = CliRunner().invoke(
        msg_container_sign_main,
        [
            "--signing-key",
            "test-signing-key",
            "--digest",
            "some-digest",
            "--reference",
            "some-reference",
            "--task-id",
            "1",
            "--config-file",
            f_config_msg_batch_signer_ok,
            "--signer-type",
            "batch",
            "--raw",
        ],
    )
    print(result.stdout)
    assert result.exit_code == 1, result.output
    assert result.output == 'no signing key "test-signing-key"\n'


def test_operations():
    signer = MsgBatchSigner()
    assert signer.operations() == [ContainerSignOperation]


def test_create_msg_batch_message(f_config_msg_batch_signer_ok):
    signer = MsgBatchSigner()
    signer.load_config(load_config(f_config_msg_batch_signer_ok))

    data = ["claim1", "claim2", "claim3"]
    with patch("uuid.uuid4", return_value="1234-5678-abcd-efgh"):
        with patch("pubtools.sign.signers.msgsigner.isodate_now") as patched_date:
            patched_date.return_value = "created-date-Z"
            operation = ContainerSignOperation(
                digests=["some-digest-1", "some-digest-2", "some-digest-3"],
                references=["some-reference-1", "some-reference-2", "some-reference-3"],
                signing_keys=["test-key"],
                task_id="1",
            )
            assert signer._create_msg_batch_message(
                data, "repo", operation, SignRequestType.BATCH
            ) == [
                MsgMessage(
                    headers={
                        "service": "pubtools-sign",
                        "environment": "prod",
                        "owner_id": "pubtools-sign-test",
                        "mtype": "batch",
                        "source": "metadata",
                    },
                    address="topic://Topic.sign",
                    body={
                        "claims": [
                            {
                                "claim_file": "claim1",
                                "sig_keynames": [""],
                                "sig_key_ids": ["test-key"],
                                "manifest_digest": "some-digest-1",
                                "repo": "repo",
                            },
                            {
                                "claim_file": "claim2",
                                "sig_keynames": [""],
                                "sig_key_ids": ["test-key"],
                                "manifest_digest": "some-digest-2",
                                "repo": "repo",
                            },
                            {
                                "claim_file": "claim3",
                                "sig_keynames": [""],
                                "sig_key_ids": ["test-key"],
                                "manifest_digest": "some-digest-3",
                                "repo": "repo",
                            },
                        ],
                        "request_id": "1234-5678-abcd-efgh",
                        "created": "created-date-Z",
                        "requested_by": "pubtools-sign-test",
                    },
                )
            ]


def test_sign(f_config_msg_batch_signer_ok):
    signer = MsgBatchSigner()
    signer.load_config(load_config(f_config_msg_batch_signer_ok))
    container_sign_operation = ContainerSignOperation(
        digests=("some-digest-1", "some-digest-2", "some-digest-3", "some-digest-4"),
        references=("some-reference-1", "some-reference-2", "some-reference-3", "some-reference-4"),
        signing_keys=["test-signing-key"],
        task_id="1",
    )

    with patch(
        "pubtools.sign.signers.msgsigner.MsgBatchSigner.container_sign"
    ) as patched_container_sign:
        signer.sign(container_sign_operation)
        patched_container_sign.assert_called_once()


@patch("uuid.uuid4", side_effect=["1234-5678-abcd-efgh", "abcd-efgh-1234-5678"])
def test_container_sign(
    patched_uuid, f_config_msg_batch_signer_ok, f_client_certificate, f_ca_certificate
):
    container_sign_operation = ContainerSignOperation(
        task_id="1",
        digests=["sha256:abcdef-1", "sha256:abcdefg-2", "sha256:abcdefg-3", "sha256:abcdefg-4"],
        references=[
            "some-registry/namespace/repo:tag1",
            "some-registry/namespace/repo:tag2",
            "some-registry/namespace/repo:tag3",
            "some-registry/namespace/repo:tag4",
        ],
        signing_keys=["test-signing-key"],
    )

    with patch("pubtools.sign.signers.msgsigner.SendClient") as patched_send_client:
        with patch("pubtools.sign.signers.msgsigner.RecvClient") as patched_recv_client:
            patched_send_client.return_value.run.return_value = []
            patched_recv_client.return_value.run.return_value = []
            patched_recv_client.return_value.recv = {
                "1234-5678-abcd-efgh": (
                    {"msg": {"errors": [], "signed_claim": "signed:'claim1'"}},
                    {"fake": "headers"},
                ),
                "abcd-efgh-1234-5678": (
                    {"msg": {"errors": [], "signed_claim": "signed:'claim2'"}},
                    {"fake": "headers"},
                ),
            }
            patched_recv_client.return_value.get_errors.return_value = []

            signer = MsgBatchSigner()
            signer.load_config(load_config(f_config_msg_batch_signer_ok))
            res = signer.container_sign(container_sign_operation)

            print(patched_send_client.mock_calls)
            patched_send_client.assert_called_with(
                messages=[ANY, ANY],
                broker_urls=["amqps://broker-01:5671", "amqps://broker-02:5671"],
                cert=f_client_certificate,
                ca_cert=f_ca_certificate,
                retries=3,
                errors=patched_recv_client.return_value.get_errors.return_value,
            )

            assert res == SigningResults(
                signer=signer,
                operation=container_sign_operation,
                signer_results=MsgSignerResults(status="ok", error_message=""),
                operation_result=ContainerSignResult(
                    results=[
                        (
                            {"msg": {"errors": [], "signed_claim": "signed:'claim1'"}},
                            {"fake": "headers"},
                        ),
                        (
                            {"msg": {"errors": [], "signed_claim": "signed:'claim2'"}},
                            {"fake": "headers"},
                        ),
                    ],
                    signing_keys=["test-signing-key"],
                    failed=False,
                ),
            )


@patch("uuid.uuid4", return_value="1234-5678-abcd-efgh")
def test_container_sign_alias(
    patched_uuid, f_config_msg_batch_signer_aliases, f_client_certificate
):
    container_sign_operation = ContainerSignOperation(
        task_id="1",
        digests=["sha256:abcdefg"],
        references=["some-registry/namespace/repo:tag"],
        signing_keys=["beta"],
    )

    with patch("pubtools.sign.signers.msgsigner.SendClient") as patched_send_client:
        with patch("pubtools.sign.signers.msgsigner.RecvClient") as patched_recv_client:
            patched_send_client.return_value.run.return_value = []
            patched_recv_client.return_value.run.return_value = []
            patched_recv_client.return_value.recv = {
                "1234-5678-abcd-efgh": (
                    {"msg": {"errors": [], "signed_claim": "signed:'claim'"}},
                    {"fake": "headers"},
                )
            }
            patched_recv_client.return_value.get_errors.return_value = []

            signer = MsgBatchSigner()
            signer.load_config(load_config(f_config_msg_batch_signer_aliases))

            with patch(
                "pubtools.sign.signers.msgsigner.MsgBatchSigner._construct_signing_batch_message"
            ) as patch_construct_signing_message:
                patch_construct_signing_message.return_value = {
                    "request_id": "1234-5678-abcd-efgh",
                }
                res = signer.container_sign(container_sign_operation)
                patch_construct_signing_message.assert_called_once_with(
                    ANY,
                    ["abcde1245"],
                    "namespace/repo",
                    extra_attrs={"pipeline_run_id": "1", "manifest_digest": ["sha256:abcdefg"]},
                    signing_key_names=[""],
                    sig_type="batch",
                )

            patched_send_client.assert_called_with(
                messages=[ANY],
                broker_urls=["amqps://broker-01:5671", "amqps://broker-02:5671"],
                cert=f_client_certificate,
                ca_cert=os.path.expanduser("~/messaging/ca-cert.crt"),
                retries=3,
                errors=patched_recv_client.return_value.get_errors.return_value,
            )

            assert res == SigningResults(
                signer=signer,
                operation=container_sign_operation,
                signer_results=MsgSignerResults(status="ok", error_message=""),
                operation_result=ContainerSignResult(
                    results=[
                        (
                            {"msg": {"errors": [], "signed_claim": "signed:'claim'"}},
                            {"fake": "headers"},
                        )
                    ],
                    signing_keys=["beta"],
                    failed=False,
                ),
            )


@patch("uuid.uuid4", return_value="1234-5678-abcd-efgh")
def test_container_sign_recv_errors(patched_uuid, f_config_msg_batch_signer_ok):
    container_sign_operation = ContainerSignOperation(
        task_id="1",
        digests=["sha256:abcdefg"],
        references=["some-registry/namespace/repo:tag"],
        signing_keys=["test-signing-key"],
    )

    with patch("pubtools.sign.signers.msgsigner.SendClient") as patched_send_client:
        with patch("pubtools.sign.signers.msgsigner.RecvClient") as patched_recv_client:
            patched_send_client.return_value.run.return_value = []
            patched_recv_client.return_value.get_errors.return_value = [
                MsgError(
                    name="TestError", description="test error description", source="test-source"
                )
            ]
            patched_recv_client.return_value.recv = {"1234-5678-abcd-efgh": "signed:'hello world'"}

            signer = MsgBatchSigner()
            signer.load_config(load_config(f_config_msg_batch_signer_ok))
            res = signer.container_sign(container_sign_operation)

            assert res == SigningResults(
                signer=signer,
                operation=container_sign_operation,
                signer_results=MsgSignerResults(
                    status="error", error_message="TestError : test error description\n"
                ),
                operation_result=ContainerSignResult(
                    results=[""], signing_keys=["test-signing-key"], failed=False
                ),
            )


@patch("uuid.uuid4", return_value="1234-5678-abcd-efgh")
def test_container_sign_send_errors(patched_uuid, f_config_msg_batch_signer_ok):
    container_sign_operation = ContainerSignOperation(
        task_id="1",
        digests=["sha256:abcdefg"],
        references=["some-registry/namespace/repo:tag"],
        signing_keys=["test-signing-key"],
    )
    with patch("pubtools.sign.signers.msgsigner.SendClient") as patched_send_client:
        with patch("pubtools.sign.signers.msgsigner.RecvClient") as patched_recv_client:
            patched_send_client.return_value.run.return_value = [
                MsgError(
                    name="TestError", description="test error description", source="test-source"
                )
            ]
            patched_recv_client.return_value.run.return_value = []
            patched_recv_client.return_value.recv = {"1234-5678-abcd-efgh": "signed:'hello world'"}

            signer = MsgBatchSigner()
            signer.load_config(load_config(f_config_msg_batch_signer_ok))
            res = signer.container_sign(container_sign_operation)

            assert res == SigningResults(
                signer=signer,
                operation=container_sign_operation,
                signer_results=MsgSignerResults(
                    status="error", error_message="TestError : test error description\n"
                ),
                operation_result=ContainerSignResult(
                    results=[""], signing_keys=["test-signing-key"], failed=False
                ),
            )


@patch("uuid.uuid4", return_value="1234-5678-abcd-efgh")
def test_container_sign_wrong_inputs(patched_uuid, f_config_msg_batch_signer_ok):
    container_sign_operation = ContainerSignOperation(
        task_id="1",
        digests=["sha256:abcdefg"],
        references=["some-registry/namespace/repo:tag", "some-registry/namespace/repo:tag2"],
        signing_keys=["test-signing-key"],
    )

    signer = MsgBatchSigner()
    signer.load_config(load_config(f_config_msg_batch_signer_ok))
    with pytest.raises(ValueError):
        signer.container_sign(container_sign_operation)


@patch("uuid.uuid4", return_value="1234-5678-abcd-efgh")
def test_container_sign_recv_timeout_fails(patched_uuid, f_config_msg_batch_signer_ok):
    container_sign_operation = ContainerSignOperation(
        task_id="1",
        digests=["sha256:abcdefg"],
        references=["some-registry/namespace/repo:tag"],
        signing_keys=["test-signing-key"],
    )

    with patch("pubtools.sign.signers.msgsigner.SendClient") as patched_send_client:
        with patch("pubtools.sign.signers.msgsigner.RecvClient") as patched_recv_client:
            patched_send_client.return_value.run.return_value = []
            patched_recv_client.return_value.get_errors.return_value = [
                MsgError(
                    name="MessagingTimeout",
                    description="Out of time when receiving messages",
                    source="test-source",
                ),
                MsgError(
                    name="MessagingTimeout",
                    description="Out of time when receiving messages",
                    source="test-source",
                ),
                MsgError(
                    name="MessagingTimeout",
                    description="Out of time when receiving messages",
                    source="test-source",
                ),
            ]
            patched_recv_client.return_value.recv = {
                "1234-5678-abcd-efgh": (
                    {"msg": {"errors": [], "signed_claim": "signed:'claim'"}},
                    {"fake": "headers"},
                )
            }

            signer = MsgBatchSigner()
            signer.load_config(load_config(f_config_msg_batch_signer_ok))
            signer.retries = 2
            signer.send_retries = 1

            res = signer.container_sign(container_sign_operation)

            assert res == SigningResults(
                signer=signer,
                operation=container_sign_operation,
                signer_results=MsgSignerResults(
                    status="error",
                    error_message="MessagingTimeout : Out of time when receiving messages\n"
                    "MessagingTimeout : Out of time when receiving messages\n",
                ),
                operation_result=ContainerSignResult(
                    results=[""], signing_keys=["test-signing-key"], failed=False
                ),
            )


@patch("uuid.uuid4", return_value="1234-5678-abcd-efgh")
def test_container_sign_recv_timeout_ok(patched_uuid, f_config_msg_batch_signer_ok):
    container_sign_operation = ContainerSignOperation(
        task_id="1",
        digests=["sha256:abcdefg"],
        references=["some-registry/namespace/repo:tag"],
        signing_keys=["test-signing-key"],
    )

    with patch("pubtools.sign.signers.msgsigner.SendClient") as patched_send_client:
        with patch("pubtools.sign.signers.msgsigner.RecvClient") as patched_recv_client:
            patched_send_client.return_value.run.return_value = []
            patched_recv_client.return_value.get_errors.return_value = [
                MsgError(
                    name="MessagingTimeout",
                    description="Out of time when receiving messages",
                    source="test-source",
                ),
            ]
            patched_recv_client.return_value.recv = {
                "1234-5678-abcd-efgh": (
                    {"msg": {"errors": [], "signed_claim": "signed:'claim'"}},
                    {"fake": "headers"},
                )
            }

            signer = MsgBatchSigner()
            signer.retries = 2
            signer.send_retries = 1

            signer.load_config(load_config(f_config_msg_batch_signer_ok))
            res = signer.container_sign(container_sign_operation)

            assert res == SigningResults(
                signer=signer,
                operation=container_sign_operation,
                signer_results=MsgSignerResults(
                    status="ok",
                    error_message="",
                ),
                operation_result=ContainerSignResult(
                    results=[
                        (
                            {"msg": {"errors": [], "signed_claim": "signed:'claim'"}},
                            {"fake": "headers"},
                        )
                    ],
                    signing_keys=["test-signing-key"],
                    failed=False,
                ),
            )


@patch("uuid.uuid4", return_value="1234-5678-abcd-efgh")
def test_clear_sign_recv_timeout(patched_uuid, f_config_msg_batch_signer_ok):
    clear_sign_operation = ClearSignOperation(
        inputs=["hello world"], signing_keys=["test-signing-key"], task_id="1", repo="repo"
    )

    with patch("pubtools.sign.signers.msgsigner.SendClient") as patched_send_client:
        with patch("pubtools.sign.signers.msgsigner.RecvClient") as patched_recv_client:
            patched_send_client.return_value.run.return_value = []
            patched_recv_client.return_value.get_errors.side_effect = [
                [
                    MsgError(
                        name="MessagingTimeout",
                        description="Out of time when receiving messages",
                        source="test-source",
                    ),
                ],
                [
                    MsgError(
                        name="MessagingTimeout",
                        description="Out of time when receiving messages",
                        source="test-source",
                    )
                ],
                [
                    MsgError(
                        name="MessagingTimeout",
                        description="Out of time when receiving messages",
                        source="test-source",
                    )
                ],
                [
                    MsgError(
                        name="MessagingTimeout",
                        description="Out of time when receiving messages",
                        source="test-source",
                    )
                ],
            ]
            patched_recv_client.return_value.recv = {"1234-5678-abcd-efgh": "signed:'hello world'"}

            signer = MsgBatchSigner()
            signer.load_config(load_config(f_config_msg_batch_signer_ok))
            signer.retries = 2
            signer.send_retries = 1

            res = signer.clear_sign(clear_sign_operation)

            assert res == SigningResults(
                signer=signer,
                operation=clear_sign_operation,
                signer_results=MsgSignerResults(
                    status="error",
                    error_message="MessagingTimeout : Out of time when receiving messages\n",
                ),
                operation_result=ClearSignResult(outputs=[""], signing_keys=["test-signing-key"]),
            )


def test_recv_client_recv_message_break(
    f_cleanup_msgsigner_messages,
    f_qpid_broker,
    f_msgsigner_listen_to_topic,
    f_fake_msgsigner,
    f_msgsigner_send_to_queue,
    f_client_certificate,
    f_ca_certificate,
    f_config_msg_batch_signer_ok2,
):
    qpid_broker, port = f_qpid_broker
    container_sign_operation = ContainerSignOperation(
        digests=("some-digest",),
        references=("some/reference:some-tag",),
        signing_keys=["test-signing-key"],
        task_id="1",
    )
    with patch(
        "pubtools.sign.clients.msg_recv_client.RecvClient.get_errors", autospec=True
    ) as patched_recv_get_errors, patch(
        "pubtools.sign.clients.msg_recv_client.RecvClient.get_received", autospec=True
    ) as patched_recv_get_received, patch(
        "uuid.uuid4", return_value="1234-5678-abcd-efgh"
    ) as _:
        patched_recv_get_errors.side_effect = [
            [
                MsgError(
                    name="MessagingTimeout",
                    description="Out of time when receiving messages",
                    source=ANY,
                )
            ],
            [
                MsgError(
                    name="MessagingTimeout",
                    description="Out of time when receiving messages",
                    source=ANY,
                )
            ],
            [
                MsgError(
                    name="MessagingTimeout",
                    description="Out of time when receiving messages",
                    source=ANY,
                )
            ],
            [
                MsgError(
                    name="MessagingTimeout",
                    description="Out of time when receiving messages",
                    source=ANY,
                )
            ],
            [],
        ]
        patched_recv_get_received.side_effect = [{"1234-5678-abcd-efgh": True}]
        signer = MsgBatchSigner()
        signer.load_config(load_config(f_config_msg_batch_signer_ok2))
        signer.retries = 2
        signer.send_retries = 1
        res = signer.container_sign(container_sign_operation)

        assert res == SigningResults(
            signer=signer,
            operation=container_sign_operation,
            signer_results=MsgSignerResults(
                status="error",
                error_message="MessagingTimeout : Out of time when receiving messages\n",
            ),
            operation_result=ContainerSignResult(
                results=[""], signing_keys=["test-signing-key"], failed=False
            ),
        )


def test_msgsig_doc_arguments():
    assert MsgBatchSigner.doc_arguments() == {
        "options": {
            "messaging_brokers": {"description": "List of brokers URLS"},
            "messaging_cert_key": {
                "description": "Client certificate + key for messaging authorization"
            },
            "messaging_ca_cert": {"description": "Messaging CA certificate"},
            "topic_send_to": {"description": "Topic where to send the messages"},
            "topic_listen_to": {"description": "Topic where to listen for replies"},
            "creator": {"description": "Identification of creator of signing request"},
            "environment": {"description": "Environment indetification in sent messages"},
            "service": {"description": "Service identificator"},
            "task_id_attribute": {
                "description": "Attribute used to custom identification of signing request"
            },
            "timeout": {"description": "Timeout for messaging receive"},
            "retries": {"description": "Retries for messaging receive"},
            "send_retries": {"description": "Retries for messaging send+receive"},
            "message_id_key": {
                "description": "Attribute name in message body which should be used as message id"
            },
            "log_level": {"description": "Log level"},
            "key_aliases": {"description": "Aliases for signing keys"},
        },
        "examples": {
            "msg_batch_signer": {
                "messaging_brokers": ["amqps://broker-01:5671", "amqps://broker-02:5671"],
                "messaging_cert_key": "~/messaging/cert.pem",
                "messaging_ca_cert": "~/messaging/ca_cert.crt",
                "topic_send_to": "topic://Topic.sign",
                "topic_listen_to": "queue://Consumer.{{creator}}.{{task_id}}.Topic.sign."
                "{{task_id}}",
                "creator": "pubtools-sign",
                "environment": "prod",
                "service": "pubtools-sign",
                "task_id_attribute": "task_id",
                "timeout": 1,
                "retries": 3,
                "send_retries": 2,
                "message_id_key": "123",
                "log_level": "debug",
                "key_aliases": "{'production':'abcde1245'}",
            }
        },
    }


def test_msgsigresult_to_dict():
    assert MsgSignerResults(status="status", error_message="error_message").to_dict() == {
        "status": "status",
        "error_message": "error_message",
    }


def test_msgsigresult_doc_arguments():
    assert MsgSignerResults.doc_arguments() == {
        "signer_result": {
            "type": "dict",
            "description": "Signing result status.",
            "returned": "always",
            "sample": {"status": "ok", "error_message": ""},
        }
    }


def test_prepare_messages_multi_repo_no_duplication(f_config_msg_batch_signer_ok):
    """Test that _prepare_messages doesn't duplicate batches across multiple repos.

    This test verifies the fix for a bug where `run_in_parallel` and `messages.extend`
    were incorrectly indented inside the repo loop, causing batch_data to be processed
    multiple times (once per repo iteration), leading to O(N²) message duplication.

    With chunk_size=2:
    - repo-a (3 digests) -> 2 batches (2 + 1)
    - repo-b (2 digests) -> 1 batch
    - repo-c (1 digest)  -> 1 batch
    Total expected: 4 batches, NOT 10 (2 + 4 + 4 from the buggy code)
    """
    signer = MsgBatchSigner()
    signer.load_config(load_config(f_config_msg_batch_signer_ok))

    # Create operation with references spanning 3 different repos
    container_sign_operation = ContainerSignOperation(
        task_id="1",
        digests=[
            # repo-a: 3 digests
            "sha256:aaa1",
            "sha256:aaa2",
            "sha256:aaa3",
            # repo-b: 2 digests
            "sha256:bbb1",
            "sha256:bbb2",
            # repo-c: 1 digest
            "sha256:ccc1",
        ],
        references=[
            # repo-a: 3 references
            "registry.example.com/namespace/repo-a:tag1",
            "registry.example.com/namespace/repo-a:tag2",
            "registry.example.com/namespace/repo-a:tag3",
            # repo-b: 2 references
            "registry.example.com/namespace/repo-b:tag1",
            "registry.example.com/namespace/repo-b:tag2",
            # repo-c: 1 reference
            "registry.example.com/namespace/repo-c:tag1",
        ],
        signing_keys=["test-signing-key"],
    )

    # Call _prepare_messages directly to test the batching logic
    with patch("pubtools.sign.signers.msgsigner.run_in_parallel") as mock_run_parallel:
        # Mock run_in_parallel to return message placeholders
        mock_run_parallel.return_value = {i: [f"msg_{i}"] for i in range(10)}

        signer._prepare_messages(container_sign_operation)

        # run_in_parallel should be called exactly ONCE with all batch_data
        assert mock_run_parallel.call_count == 1, (
            f"run_in_parallel should be called once, but was called "
            f"{mock_run_parallel.call_count} times. This indicates the bug where "
            f"run_in_parallel was incorrectly inside the repo loop."
        )

        # Verify the batch_data passed to run_in_parallel
        call_args = mock_run_parallel.call_args
        batch_data = list(call_args[0][1])  # Second positional arg is the data iterable

        # With chunk_size=2:
        # - repo-a (3 digests): 2 batches (chunk of 2, chunk of 1)
        # - repo-b (2 digests): 1 batch (chunk of 2)
        # - repo-c (1 digest): 1 batch (chunk of 1)
        # Total: 4 batches
        expected_batch_count = 4
        assert len(batch_data) == expected_batch_count, (
            f"Expected {expected_batch_count} batches, got {len(batch_data)}. "
            f"Batch data: {[fd.args[1] for fd in batch_data]}"  # Show repo names
        )

        # Verify each repo appears the correct number of times
        repo_counts = {}
        for fdata in batch_data:
            repo = fdata.args[1]  # repo is the second arg
            repo_counts[repo] = repo_counts.get(repo, 0) + 1

        assert repo_counts.get("namespace/repo-a") == 2, "repo-a should have 2 batches"
        assert repo_counts.get("namespace/repo-b") == 1, "repo-b should have 1 batch"
        assert repo_counts.get("namespace/repo-c") == 1, "repo-c should have 1 batch"



from click.testing import CliRunner
import pytest
import requests_mock
from unittest.mock import patch, Mock, call, ANY

from pubtools.sign.signers.cosignsigner import (
    CosignSigner,
    CosignSignerResults,
    ContainerSignOperation,
    ContainerSignResult,
    cosign_container_sign_main,
    cosign_list_existing_signatures,
)
from pubtools.sign.conf.conf import load_config
from pubtools.sign.exceptions import UnsupportedOperation
from pubtools.sign.results.signing_results import SigningResults


def mock_registry_responses(requests_mock, registry, repo, sig_response=200, registry_response=401):
    requests_mock.get(  # nosec
        "https://example-registry.io/v2/auth?service=example-registry.io"
        "&scope=repository%3Anamespace%2Frepo%3Apull",
        [
            {
                "headers": {"authorization": "Bearer some-token"},
                "status_code": 200,
                "json": {},
            },
        ],
    )
    requests_mock.get(  # nosec
        "https://example-registry.io/v2/namespace/repo/manifests/sha256-abcdefg.sig",
        [
            {
                "headers": {
                    "www-authenticate": f'Bearer realm="https://{registry}/v2/auth",'
                    f'service="{registry}",scope="repository:{repo}:pull"'
                },
                "status_code": registry_response,
            },
            {
                "headers": {
                    "www-authenticate": f'Bearer realm="https://{registry}/v2/auth",'
                    f'service="{registry}",scope="repository:{repo}:pull'
                },
                "status_code": sig_response,
                "json": {},
            },
        ],
    )


@pytest.fixture
def f_expected_container_sign_args(f_config_cosign_signer_ok):
    return [
        "--signing-key",
        "test-signing-key",
        "--digest",
        "some-digest",
        "--reference",
        "some-reference",
        "--config-file",
        f_config_cosign_signer_ok,
    ]


@pytest.fixture
def f_expected_container_sign_identity_args(f_config_cosign_signer_ok):
    return [
        "--signing-key",
        "test-signing-key",
        "--digest",
        "some-digest",
        "--reference",
        "some-reference",
        "--identity",
        "some-registry/namespace/repo",
        "--config-file",
        f_config_cosign_signer_ok,
    ]


@pytest.fixture
def f_expected_cosign_sign_args():
    return [
        "/usr/bin/cosign",
        "-t",
        "30s",
        "sign",
        "-y",
        "--key",
        "test-signing-key",
        "--allow-http-registry=false",
        "--allow-insecure-registry=false",
        "--rekor-url",
        "https://rekor.sigstore.dev",
        "--tlog-upload=true",
        "--registry-username",
        "some-user",
        "--registry-password",
        "some-password",
        "-a",
        "tag=tag",
        "some-registry/namespace/repo@sha256:abcdefg",
    ]


@pytest.fixture
def f_expected_cosign_sign_identity_args():
    cosign_args = []
    for tag in ["tag1", "tag2"]:
        args = [
            "/usr/bin/cosign",
            "-t",
            "30s",
            "sign",
            "-y",
            "--key",
            "test-signing-key",
            "--allow-http-registry=false",
            "--allow-insecure-registry=false",
            "--rekor-url",
            "https://rekor.sigstore.dev",
            "--tlog-upload=true",
            "--registry-username",
            "some-user",
            "--registry-password",
            "some-password",
            "--sign-container-identity",
            "some-registry/namespace/repo",
            "-a",
            f"tag={tag}",
            "some-registry/namespace/repo@sha256:abcdefg",
        ]
        cosign_args.append(args)
    return cosign_args


@pytest.fixture
def f_expected_cosign_triangulate_args():
    return [
        "/usr/bin/cosign",
        "-t",
        "30s",
        "triangulate",
        "--allow-http-registry=false",
        "--allow-insecure-registry=false",
        "--registry-username",
        "some-user",
        "--registry-password",
        "some-password",
        "example-registry.io/namespace/repo:latest",
    ]


def test_cosign_container_sign(f_cosign_signer, f_expected_container_sign_args):
    f_cosign_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "ok"
    }
    f_cosign_signer.return_value.sign.return_value.operation_result.results = []
    f_cosign_signer.return_value.sign.return_value.operation_result.signing_keys = []
    f_cosign_signer.return_value.sign.return_value.operation.to_dict.return_value = {}
    result = CliRunner().invoke(cosign_container_sign_main, f_expected_container_sign_args)
    print(result.stdout)
    assert result.exit_code == 0, result.output


def test_cosign_container_identity_sign(f_cosign_signer, f_expected_container_sign_identity_args):
    f_cosign_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "ok"
    }
    f_cosign_signer.return_value.sign.return_value.operation_result.results = []
    f_cosign_signer.return_value.sign.return_value.operation_result.signing_keys = []
    f_cosign_signer.return_value.sign.return_value.operation.to_dict.return_value = {}
    result = CliRunner().invoke(cosign_container_sign_main, f_expected_container_sign_identity_args)
    print(result.stdout)
    assert result.exit_code == 0, result.output


def test_cosign_container_sign_error(f_cosign_signer, f_expected_container_sign_args):
    f_cosign_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "error",
        "error_message": "simulated error",
    }
    f_cosign_signer.return_value.sign.return_value.operation_result.results = []
    f_cosign_signer.return_value.sign.return_value.operation_result.signing_keys = []
    f_cosign_signer.return_value.sign.return_value.operation.to_dict.return_value = {}
    result = CliRunner().invoke(
        cosign_container_sign_main,
        f_expected_container_sign_args,
    )
    print(result.stdout)
    assert result.exit_code == 1, result.output


def test_cosign_container_sign_raw(f_cosign_signer, f_expected_container_sign_args):
    f_cosign_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "ok"
    }
    f_cosign_signer.return_value.sign.return_value.operation_result.results = ["signed"]
    f_cosign_signer.return_value.sign.return_value.operation_result.signing_keys = []
    result = CliRunner().invoke(
        cosign_container_sign_main, f_expected_container_sign_args + ["--raw"]
    )
    print(result.stdout)
    assert result.exit_code == 0, result.output
    assert result.output == "signed\n"


def test_cosign_container_sign_raw_error(f_cosign_signer, f_expected_container_sign_args):
    f_cosign_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "error",
        "error_message": "simulated error",
    }
    f_cosign_signer.return_value.sign.return_value.operation_result.results = []
    f_cosign_signer.return_value.sign.return_value.operation_result.signing_keys = []
    result = CliRunner().invoke(
        cosign_container_sign_main, f_expected_container_sign_args + ["--raw"]
    )
    print(result.stdout)
    assert result.exit_code == 1, result.output
    assert result.output == "simulated error\n"


def test_operations():
    signer = CosignSigner()
    assert signer.operations() == [ContainerSignOperation]


def test_sign(f_config_cosign_signer_ok):
    signer = CosignSigner()
    signer.load_config(load_config(f_config_cosign_signer_ok))
    container_sign_operation = ContainerSignOperation(
        digests=("some-digest",),
        references=("some-reference",),
        signing_keys=["test-signing-key"],
        task_id="",
    )
    with patch(
        "pubtools.sign.signers.cosignsigner.CosignSigner.container_sign"
    ) as patched_container_sign:
        signer.sign(container_sign_operation)
        patched_container_sign.assert_called_once()
        with pytest.raises(UnsupportedOperation):
            signer.sign(Mock())


def test_container_sign(f_config_cosign_signer_ok, f_environ, f_expected_cosign_sign_args):
    container_sign_operation = ContainerSignOperation(
        task_id="",
        digests=["sha256:abcdefg"],
        references=["some-registry/namespace/repo:tag"],
        signing_keys=["test-signing-key"],
    )

    with patch("subprocess.Popen") as patched_popen:
        patched_popen().returncode = 0
        patched_popen().communicate.return_value = ("stdout", "stderr")

        signer = CosignSigner()
        signer.load_config(load_config(f_config_cosign_signer_ok))
        res = signer.container_sign(container_sign_operation)

        patched_popen.assert_has_calls(
            [
                call(
                    f_expected_cosign_sign_args,
                    env={"PYTEST_CURRENT_TEST": ANY},
                    stderr=-1,
                    stdout=-1,
                    text=True,
                )
            ]
        )

        assert res == SigningResults(
            signer=signer,
            operation=container_sign_operation,
            signer_results=CosignSignerResults(status="ok", error_message=""),
            operation_result=ContainerSignResult(
                results=["stderr"], signing_keys=["test-signing-key"], failed=False
            ),
        )


def test_container_sign_identity(
    f_config_cosign_signer_ok, f_environ, f_expected_cosign_sign_identity_args
):
    container_sign_operation = ContainerSignOperation(
        task_id="",
        digests=["sha256:abcdefg", "sha256:abcdefg"],
        references=["some-registry/namespace/repo:tag1", "some-registry/namespace/repo:tag2"],
        identity_references=["some-registry/namespace/repo", "some-registry/namespace/repo"],
        signing_keys=["test-signing-key"],
    )

    with patch("subprocess.Popen") as patched_popen:
        patched_popen().returncode = 0
        patched_popen().communicate.return_value = ("stdout", "stderr")

        signer = CosignSigner()
        signer.load_config(load_config(f_config_cosign_signer_ok))
        res = signer.container_sign(container_sign_operation)

        patched_popen.assert_has_calls(
            [
                call(
                    f_expected_cosign_sign_identity_args[0],
                    env={"PYTEST_CURRENT_TEST": ANY},
                    stderr=-1,
                    stdout=-1,
                    text=True,
                )
            ]
        )
        patched_popen.assert_has_calls(
            [
                call(
                    f_expected_cosign_sign_identity_args[1],
                    env={"PYTEST_CURRENT_TEST": ANY},
                    stderr=-1,
                    stdout=-1,
                    text=True,
                )
            ]
        )

        assert res == SigningResults(
            signer=signer,
            operation=container_sign_operation,
            signer_results=CosignSignerResults(status="ok", error_message=""),
            operation_result=ContainerSignResult(
                results=["stderr", "stderr"], signing_keys=["test-signing-key"], failed=False
            ),
        )


def test_container_sign_alias(f_config_cosign_signer_aliases, f_environ):
    container_sign_operation = ContainerSignOperation(
        task_id="",
        digests=["sha256:abcdefg"],
        references=["some-registry/namespace/repo:tag"],
        signing_keys=["beta"],
    )

    with patch("subprocess.Popen") as patched_popen:
        patched_popen().returncode = 0
        patched_popen().communicate.return_value = ("stdout", "stderr")

        signer = CosignSigner()
        signer.load_config(load_config(f_config_cosign_signer_aliases))
        res = signer.container_sign(container_sign_operation)

        patched_popen.assert_has_calls(
            [
                call(
                    [
                        "/usr/bin/cosign",
                        "-t",
                        "30s",
                        "sign",
                        "-y",
                        "--key",
                        "abcde1245",
                        "--allow-http-registry=false",
                        "--allow-insecure-registry=false",
                        "--rekor-url",
                        "https://rekor.sigstore.dev",
                        "--tlog-upload=true",
                        "-a",
                        "tag=tag",
                        "some-registry/namespace/repo@sha256:abcdefg",
                    ],
                    env={"PYTEST_CURRENT_TEST": ANY},
                    stderr=-1,
                    stdout=-1,
                    text=True,
                )
            ]
        )

        assert res == SigningResults(
            signer=signer,
            operation=container_sign_operation,
            signer_results=CosignSignerResults(status="ok", error_message=""),
            operation_result=ContainerSignResult(
                results=["stderr"], signing_keys=["beta"], failed=False
            ),
        )


def test_container_sign_error(f_config_cosign_signer_ok, f_environ, f_expected_cosign_sign_args):
    container_sign_operation = ContainerSignOperation(
        task_id="",
        digests=["sha256:abcdefg"],
        references=["some-registry/namespace/repo:tag"],
        signing_keys=["test-signing-key"],
    )

    with patch("subprocess.Popen") as patched_popen:
        patched_popen().returncode = 1
        patched_popen().communicate.return_value = ("stdout", "stderr")

        signer = CosignSigner()
        signer.load_config(load_config(f_config_cosign_signer_ok))
        res = signer.container_sign(container_sign_operation)

        patched_popen.assert_has_calls(
            [
                call(
                    f_expected_cosign_sign_args,
                    env={"PYTEST_CURRENT_TEST": ANY},
                    stderr=-1,
                    stdout=-1,
                    text=True,
                )
            ]
        )

        assert res == SigningResults(
            signer=signer,
            operation=container_sign_operation,
            signer_results=CosignSignerResults(status="failed", error_message="stderr"),
            operation_result=ContainerSignResult(
                results=["stderr"], signing_keys=["test-signing-key"], failed=True
            ),
        )


def test_container_sign_digests_only(
    f_config_cosign_signer_ok, f_environ, f_expected_cosign_sign_args
):
    container_sign_operation = ContainerSignOperation(
        task_id="",
        digests=["some-registry/namespace/repo@sha256:abcdefg"],
        references=[],
        signing_keys=["test-signing-key"],
    )

    with patch("subprocess.Popen") as patched_popen:
        patched_popen().returncode = 0
        patched_popen().communicate.return_value = ("stdout", "stderr")

        signer = CosignSigner()
        signer.load_config(load_config(f_config_cosign_signer_ok))
        res = signer.container_sign(container_sign_operation)

        patched_popen.assert_has_calls(
            [
                call(
                    [
                        "/usr/bin/cosign",
                        "-t",
                        "30s",
                        "sign",
                        "-y",
                        "--key",
                        "test-signing-key",
                        "--allow-http-registry=false",
                        "--allow-insecure-registry=false",
                        "--rekor-url",
                        "https://rekor.sigstore.dev",
                        "--tlog-upload=true",
                        "--registry-username",
                        "some-user",
                        "--registry-password",
                        "some-password",
                        "some-registry/namespace/repo@sha256:abcdefg",
                    ],
                    env=ANY,
                    stderr=-1,
                    stdout=-1,
                    text=True,
                )
            ]
        )

        assert res == SigningResults(
            signer=signer,
            operation=container_sign_operation,
            signer_results=CosignSignerResults(status="ok", error_message=""),
            operation_result=ContainerSignResult(
                results=["stderr"], signing_keys=["test-signing-key"], failed=False
            ),
        )


def test_container_sign_digests_only_indentity(
    f_config_cosign_signer_ok, f_environ, f_expected_cosign_sign_args
):
    container_sign_operation = ContainerSignOperation(
        task_id="",
        digests=["some-registry/namespace/repo@sha256:abcdefg"],
        references=[],
        identity_references=["some-registry/namespace/repo"],
        signing_keys=["test-signing-key"],
    )

    with patch("subprocess.Popen") as patched_popen:
        patched_popen().returncode = 0
        patched_popen().communicate.return_value = ("stdout", "stderr")

        signer = CosignSigner()
        signer.load_config(load_config(f_config_cosign_signer_ok))
        res = signer.container_sign(container_sign_operation)

        patched_popen.assert_has_calls(
            [
                call(
                    [
                        "/usr/bin/cosign",
                        "-t",
                        "30s",
                        "sign",
                        "-y",
                        "--key",
                        "test-signing-key",
                        "--allow-http-registry=false",
                        "--allow-insecure-registry=false",
                        "--rekor-url",
                        "https://rekor.sigstore.dev",
                        "--tlog-upload=true",
                        "--registry-username",
                        "some-user",
                        "--registry-password",
                        "some-password",
                        "--sign-container-identity",
                        "some-registry/namespace/repo",
                        "some-registry/namespace/repo@sha256:abcdefg",
                    ],
                    env=ANY,
                    stderr=-1,
                    stdout=-1,
                    text=True,
                )
            ]
        )

        assert res == SigningResults(
            signer=signer,
            operation=container_sign_operation,
            signer_results=CosignSignerResults(status="ok", error_message=""),
            operation_result=ContainerSignResult(
                results=["stderr"], signing_keys=["test-signing-key"], failed=False
            ),
        )


def test_container_sign_mismatch_refs(f_config_cosign_signer_ok):
    container_sign_operation = ContainerSignOperation(
        task_id="",
        digests=["sha256:abcdefg"],
        references=["some-registry/namespace/repo:tag1", "some-registry/namespace/repo:tag2"],
        signing_keys=["test-signing-key"],
    )

    with patch("subprocess.Popen") as patched_popen:
        signer = CosignSigner()
        signer.load_config(load_config(f_config_cosign_signer_ok))
        with pytest.raises(ValueError):
            signer.container_sign(container_sign_operation)

        print(patched_popen.mock_calls)
        patched_popen.assert_not_called()


def test_cosignsig_doc_arguments():
    assert CosignSigner.doc_arguments() == {
        "options": {
            "log_level": {"description": "Log level"},
            "registry_auth_file": {"description": "Registry basic auth file"},
            "registry_password": {"description": "Registry basic password"},
            "registry_user": {"description": "Registry basic user"},
            "rekor_url": {"description": "URL for rekor stl server"},
            "timeout": {"description": "Timeout for cosign operations with units"},
            "upload_tlog": {"description": "upload signing record to rekor"},
            "cosign_bin": {"description": "Path to cosign binary"},
            "allow_http_registry": {"description": "Allow http registry"},
            "allow_insecure_registry": {"description": "Allow insecure registry"},
            "env_variables": {"description": "environment variables used for signing"},
            "key_aliases": {"description": "Aliases for signing keys"},
            "retries": {"description": "Number of retries for http requests"},
            "num_threads": {"description": "The number of threads for running cosign command"},
        },
        "examples": {
            "cosign_signer": {
                "log_level": "debug",
                "registry_auth_file": "auth.json",
                "registry_password": "password",
                "registry_user": "username",
                "allow_http_registry": False,
                "allow_insecure_registry": False,
                "cosign_bin": "/usr/local/bin/cosign",
                "env_variables": '{"AWS_ACCESS_KEY_ID": "xxxxxxx",'
                ' "AWS_SECRET_ACCESS_KEY":"yyyyyyyyy"}',
                "rekor_url": "https://rekor.sigstore.dev",
                "timeout": "60s",
                "upload_tlog": "False",
                "key_aliases": "{'production':'abcde1245'}",
                "retries": 5,
                "num_threads": 10,
            }
        },
    }


def test_cosignsigresult_to_dict():
    assert CosignSignerResults(status="status", error_message="error_message").to_dict() == {
        "status": "status",
        "error_message": "error_message",
    }


def test_msgsigresult_doc_arguments():
    assert CosignSignerResults.doc_arguments() == {
        "signer_result": {
            "type": "dict",
            "description": "Signing result status.",
            "returned": "always",
            "sample": {"status": "ok", "error_message": ""},
        }
    }


def test_container_existing_signatures(
    f_config_cosign_signer_ok, f_environ, f_expected_cosign_triangulate_args
):
    with patch("subprocess.Popen") as patched_popen:
        patched_popen().returncode = 0
        patched_popen().communicate.return_value = (
            "example-registry.io/namespace/repo:sha256-abcdefg.sig",
            "stderr",
        )

        signer = CosignSigner()
        signer.load_config(load_config(f_config_cosign_signer_ok))

        with requests_mock.Mocker() as m:
            mock_registry_responses(m, "example-registry.io", "namespace/repo")
            res = signer.existing_signatures("example-registry.io/namespace/repo:latest")

        patched_popen.assert_has_calls(
            [
                call(
                    f_expected_cosign_triangulate_args,
                    env=ANY,
                    stderr=-1,
                    stdout=-1,
                    text=True,
                )
            ]
        )
        assert res == (True, ["example-registry.io/namespace/repo:sha256-abcdefg.sig"])


def test_container_existing_signatures_error(
    f_config_cosign_signer_ok, f_environ, f_expected_cosign_triangulate_args
):
    with patch("subprocess.Popen") as patched_popen:
        patched_popen().returncode = 1
        patched_popen().communicate.return_value = ("stdout1\nstdout2", "stderr")

        signer = CosignSigner()
        signer.load_config(load_config(f_config_cosign_signer_ok))

        with requests_mock.Mocker() as m:
            mock_registry_responses(m, "example-registry.io", "namespace/repo")
            res = signer.existing_signatures("example-registry.io/namespace/repo:latest")

        patched_popen.assert_has_calls(
            [
                call(
                    f_expected_cosign_triangulate_args,
                    env=ANY,
                    stderr=-1,
                    stdout=-1,
                    text=True,
                )
            ]
        )
        assert res == (False, "stderr")


def test_container_existing_signatures_no_signature(
    f_config_cosign_signer_ok, f_environ, f_expected_cosign_triangulate_args
):
    with patch("subprocess.Popen") as patched_popen:
        patched_popen().returncode = 0
        patched_popen().communicate.return_value = (
            "example-registry.io/namespace/repo:sha256-abcdefg.sig",
            "stderr",
        )
        signer = CosignSigner()
        signer.load_config(load_config(f_config_cosign_signer_ok))

        with requests_mock.Mocker() as m:
            mock_registry_responses(m, "example-registry.io", "namespace/repo", sig_response=404)
            res = signer.existing_signatures("example-registry.io/namespace/repo:latest")

        patched_popen.assert_has_calls(
            [
                call(
                    f_expected_cosign_triangulate_args,
                    env=ANY,
                    stderr=-1,
                    stdout=-1,
                    text=True,
                )
            ]
        )
        assert res == (True, "")


def test_container_existing_signatures_no_auth_needed(
    f_config_cosign_signer_ok, f_environ, f_expected_cosign_triangulate_args
):
    with patch("subprocess.Popen") as patched_popen:
        patched_popen().returncode = 0
        patched_popen().communicate.return_value = (
            "example-registry.io/namespace/repo:sha256-abcdefg.sig",
            "stderr",
        )
        signer = CosignSigner()
        signer.load_config(load_config(f_config_cosign_signer_ok))

        with requests_mock.Mocker() as m:
            mock_registry_responses(
                m, "example-registry.io", "namespace/repo", registry_response=200
            )
            res = signer.existing_signatures("example-registry.io/namespace/repo:latest")

        patched_popen.assert_has_calls(
            [
                call(
                    f_expected_cosign_triangulate_args,
                    env=ANY,
                    stderr=-1,
                    stdout=-1,
                    text=True,
                )
            ]
        )
        assert res == (True, ["example-registry.io/namespace/repo:sha256-abcdefg.sig"])


def test_container_existing_signatures_no_auth_provided(
    f_config_cosign_signer_no_auth, f_environ, f_expected_cosign_triangulate_args
):
    with patch("subprocess.Popen") as patched_popen:
        patched_popen().returncode = 0
        patched_popen().communicate.return_value = (
            "example-registry.io/namespace/repo:sha256-abcdefg.sig",
            "stderr",
        )
        signer = CosignSigner()
        signer.load_config(load_config(f_config_cosign_signer_no_auth))

        with requests_mock.Mocker() as m:
            mock_registry_responses(
                m, "example-registry.io", "namespace/repo", registry_response=401
            )
            with pytest.raises(ValueError):
                signer.existing_signatures("example-registry.io/namespace/repo:latest")


def test_container_existing_signatures_repo_no_found(
    f_config_cosign_signer_ok, f_environ, f_expected_cosign_triangulate_args
):
    with patch("subprocess.Popen") as patched_popen:
        patched_popen().returncode = 0
        patched_popen().communicate.return_value = (
            "example-registry.io/namespace/repo:sha256-abcdefg.sig",
            "stderr",
        )
        signer = CosignSigner()
        signer.load_config(load_config(f_config_cosign_signer_ok))

        with requests_mock.Mocker() as m:
            mock_registry_responses(
                m,
                "example-registry.io",
                "namespace/repo",
                sig_response=200,
                registry_response=404,
            )
            res = signer.existing_signatures("example-registry.io/namespace/repo:latest")

        patched_popen.assert_has_calls(
            [
                call(
                    f_expected_cosign_triangulate_args,
                    env=ANY,
                    stderr=-1,
                    stdout=-1,
                    text=True,
                )
            ]
        )
        assert res == (True, "")


def test_container_existing_signatures_repo_registry_sig_error(
    f_config_cosign_signer_ok, f_environ, f_expected_cosign_triangulate_args
):
    with patch("subprocess.Popen") as patched_popen:
        patched_popen().returncode = 0
        patched_popen().communicate.return_value = (
            "example-registry.io/namespace/repo:sha256-abcdefg.sig",
            "stderr",
        )
        signer = CosignSigner()
        signer.load_config(load_config(f_config_cosign_signer_ok))

        with requests_mock.Mocker() as m:
            mock_registry_responses(
                m,
                "example-registry.io",
                "namespace/repo",
                sig_response=500,
            )
            res = signer.existing_signatures("example-registry.io/namespace/repo:latest")

        patched_popen.assert_has_calls(
            [
                call(
                    f_expected_cosign_triangulate_args,
                    env=ANY,
                    stderr=-1,
                    stdout=-1,
                    text=True,
                )
            ]
        )
        assert res == (False, "Unexpected Error: 500 - {}")


def test_container_existing_signatures_repo_registry_registry_error(
    f_config_cosign_signer_ok, f_environ, f_expected_cosign_triangulate_args
):
    with patch("subprocess.Popen") as patched_popen:
        patched_popen().returncode = 0
        patched_popen().communicate.return_value = (
            "example-registry.io/namespace/repo:sha256-abcdefg.sig",
            "stderr",
        )
        signer = CosignSigner()
        signer.load_config(load_config(f_config_cosign_signer_ok))

        with requests_mock.Mocker() as m:
            mock_registry_responses(
                m,
                "example-registry.io",
                "namespace/repo",
                registry_response=500,
            )
            res = signer.existing_signatures("example-registry.io/namespace/repo:latest")

        patched_popen.assert_has_calls(
            [
                call(
                    f_expected_cosign_triangulate_args,
                    env=ANY,
                    stderr=-1,
                    stdout=-1,
                    text=True,
                )
            ]
        )
        assert res == (False, "Unexpected Error: 500 - ")


def test_container_existing_signatures_main(
    f_config_cosign_signer_ok, f_environ, f_expected_cosign_triangulate_args
):
    with patch("subprocess.Popen") as patched_popen:
        patched_popen().returncode = 0
        patched_popen().communicate.return_value = (
            "example-registry.io/namespace/repo:sha256-abcdefg.sig",
            "stderr",
        )
        with requests_mock.Mocker() as m:
            mock_registry_responses(m, "example-registry.io", "namespace/repo")
            res = cosign_list_existing_signatures(
                f_config_cosign_signer_ok, "example-registry.io/namespace/repo:latest"
            )

        patched_popen.assert_has_calls(
            [
                call(
                    f_expected_cosign_triangulate_args,
                    env=ANY,
                    stderr=-1,
                    stdout=-1,
                    text=True,
                )
            ]
        )
        assert res == (True, ["example-registry.io/namespace/repo:sha256-abcdefg.sig"])



import dataclasses
from typing import Dict, Any, Optional


@dataclasses.dataclass
class MsgMessage:
    """Messaging message model.

    Attributes:
        headers (Dict[str, Any]): Headers of the message.
        address (str): Address to which the message is sent.
        body (Dict[str, Any]): Body of the message.
    """

    headers: Dict[str, Any]
    address: str
    body: Dict[str, Any]


@dataclasses.dataclass
class MsgError:
    """Messaging error model.

    Attributes:
        name (str): Name of the error.
        description (Optional[str]): Description of the error.
        source (Any): Source of the error, can be a link, session, connection, or transport.
    """

    name: str
    description: Optional[str]
    source: Any



import json
import logging
from typing import List, Dict, Any

from ..models.msg import MsgMessage, MsgError

from .msg import _MsgClient

import proton
import proton.utils
from proton.reactor import Container

from pubtools.tracing import get_trace_wrapper

from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator

tw = get_trace_wrapper()
propagator = TraceContextTextMapPropagator()
LOG = logging.getLogger("pubtools.sign.signers.msg_send_client")


class _SendClient(_MsgClient):
    def __init__(
        self,
        messages: List[MsgMessage],
        broker_urls: List[str],
        cert: str,
        ca_cert: str,
        errors: List[MsgError],
        **kwargs: Dict[str, Any],
    ) -> None:
        super().__init__(errors=errors, **kwargs)
        self.broker_urls = broker_urls
        self.messages = messages
        self.ssl_domain = proton.SSLDomain(proton.SSLDomain.MODE_CLIENT)
        if cert:
            LOG.info(self._format_log_msg(f"SENDER: Using {cert} as SSL certificate"))
            self.ssl_domain.set_credentials(cert, cert, None)
        if ca_cert:
            LOG.info(self._format_log_msg(f"SENDER: Using {ca_cert} as SSL CA certificate"))
            self.ssl_domain.set_trusted_ca_db(ca_cert)
        self.ssl_domain.set_peer_authentication(proton.SSLDomain.ANONYMOUS_PEER)
        self.sent = 0
        self.confirmed = 0
        self.total = len(messages)

    def on_start(self, event: proton.Event) -> None:
        event.container.connect(
            urls=self.broker_urls, ssl_domain=self.ssl_domain, sasl_enabled=False
        )

    def on_connection_opened(self, event: proton.Event) -> None:
        if event.connection:
            self.sender = event.container.create_sender(event.connection)

    @tw.instrument_func()
    def on_sendable(self, event: proton.Event) -> None:
        if event.sender and event.sender.credit and self.sent < self.total:
            message = self.messages[self.sent]
            # Inject trace context to message properties
            propagator.inject(carrier=message.headers)
            LOG.debug(
                self._format_log_msg(
                    f"SENDER: Sending message: {json.dumps(message.body)}"
                    f"{message.address}"
                    f"{json.dumps(message.headers)}",
                )
            )
            if event.sender:
                event.sender.send(
                    proton.Message(
                        properties=message.headers,
                        address=message.address,
                        body=json.dumps(message.body),
                    )
                )
            self.sent += 1

    def on_accepted(self, event: proton.Event) -> None:
        # LOG.info("Sender accepted")
        self.confirmed += 1
        if self.confirmed == self.total:
            LOG.info("SENDER: closing")
            if event.connection:
                event.connection.close()
            self.sender.close()

    def on_disconnected(self, event: proton.Event) -> None:  # pragma: no cover
        self.sent = self.confirmed  # pragma: no cover

    def on_error(self, event: proton.Event, source: Any = None) -> bool:
        not_ignored = super().on_error(event, source)
        if not not_ignored:
            LOG.warning(
                self._format_log_msg(
                    f"SENDER Error: {source.condition or source.remote_condition} ignored",
                    event=event,
                )
            )
        else:
            LOG.error(
                self._format_log_msg(
                    f"SENDER Error: {source.condition or source.remote_condition}", event=event
                )
            )
        return not_ignored


class SendClient(Container):
    """SendClient wrapper class."""

    def __init__(
        self,
        messages: List[MsgMessage],
        broker_urls: List[str],
        cert: str,
        ca_cert: str,
        retries: int,
        errors: List[MsgError],
        **kwargs: Dict[str, Any],
    ) -> None:
        """Send Client Initializer.

        Args:
            messages (List[MsgMessage]): List of messages to send.
            broker_urls (List[str]): List of addresses of messaging broker.
            cert (str): Path to messaging client key and certificate in PEM format.
            ca_cert (str): Path to messaging CA certificate.
            retries (int): Number of retries for sending messages.
            errors (List[MsgError]): List of errors which occurred during the process.
        """
        self.messages = messages
        self.handler = _SendClient(
            messages=messages, broker_urls=broker_urls, cert=cert, ca_cert=ca_cert, errors=errors
        )
        self._retries = retries
        self._errors = errors
        super().__init__(self.handler, **kwargs)

    def run(self) -> List[MsgError]:  # type: ignore[override]
        """Run the SendClient.

        Returns:
            List[MsgError]: List of errors that occurred during sending messages.
        """
        errors_len = 0
        if not len(self.messages):
            LOG.warning("No messages to send")
            return []
        for x in range(self._retries):
            super().run()
            if len(self._errors) == errors_len:
                break
            errors_len = len(self._errors)
        else:
            return self._errors
        return []



from __future__ import annotations

from abc import ABC, abstractmethod
import dataclasses
from typing import Any, Dict, List


@dataclasses.dataclass()
class OperationResult(ABC):
    """OperationResult abstract class."""

    signing_keys: List[str]

    @abstractmethod
    def to_dict(self: OperationResult) -> Dict[Any, Any]:
        """Return dict representation of OperationResult."""
        ...  # pragma: no cover



from __future__ import annotations

from dataclasses import field, dataclass

from typing import List, ClassVar, Any

from ..results.operation_result import OperationResult

from .base import SignOperation


@dataclass
class ClearSignOperation(SignOperation):
    """ClearsSignOperation model class."""

    ResultType: ClassVar[OperationResult]
    inputs: List[str] = field(
        metadata={
            "type": "list",
            "description": "Signing data",
            "required": "true",
            "sample": ["input1", "input2"],
        },
        default_factory=list,
    )
    task_id: str = field(
        metadata={
            "type": "str",
            "description": "Usually pub task id, serves as identifier for in signing request",
            "required": "true",
            "sample": "1",
        },
        default="",
    )
    repo: str = field(
        metadata={
            "type": "str",
            "description": "Repository name",
            "required": "true",
            "sample": "repo",
        },
        default="",
    )
    requester: str = field(
        metadata={"description": "Requester of the signing operation"}, default=""
    )

    def to_dict(self) -> dict[str, Any]:
        """Return a dict representation of the object."""
        return dict(
            inputs=self.inputs,
            signing_keys=self.signing_keys,
            task_id=self.task_id,
            repo=self.repo,
            requester=self.requester,
        )



from __future__ import annotations

from dataclasses import field, dataclass
import itertools
import json
import logging
from typing import Dict, List, ClassVar, Any, Tuple, Type, Optional
from typing_extensions import Self
import os
import sys

import click

from . import Signer
from ..operations.base import SignOperation
from ..operations import ContainerSignOperation
from ..results.signing_results import SigningResults
from ..results import ContainerSignResult
from ..results import SignerResults
from ..exceptions import UnsupportedOperation
from ..conf.conf import load_config, CONFIG_PATHS
from ..utils import set_log_level, run_command, _get_config_file, run_in_parallel, FData
from ..clients.registry import ContainerRegistryClient, AuthTokenWrapper


LOG = logging.getLogger("pubtools.sign.signers.cosignsigner")


@dataclass()
class CosignSignerResults(SignerResults):
    """CosignSignerResults model."""

    status: str
    error_message: str

    def to_dict(self: SignerResults) -> Dict[str, Any]:
        """Return dict representation of MsgSignerResults model."""
        return {"status": self.status, "error_message": self.error_message}

    @classmethod
    def doc_arguments(cls: Type[Self]) -> Dict[str, Any]:
        """Return dictionary with result description of SignerResults."""
        doc_arguments = {
            "signer_result": {
                "type": "dict",
                "description": "Signing result status.",
                "returned": "always",
                "sample": {"status": "ok", "error_message": ""},
            }
        }

        return doc_arguments


@dataclass()
class CosignSigner(Signer):
    """Cosign signer class."""

    cosign_bin: str = field(
        init=False,
        metadata={
            "description": "Path to cosign binary",
            "sample": "/usr/local/bin/cosign",
        },
        default="/usr/bin/cosign",
    )

    timeout: str = field(
        init=False,
        metadata={
            "description": "Timeout for cosign operations with units",
            "sample": "60s",
        },
        default="3m0s",
    )

    num_threads: int = field(
        init=False,
        metadata={
            "description": "The number of threads for running cosign command",
            "sample": 10,
        },
        default=10,
    )
    allow_http_registry: bool = field(
        init=False,
        metadata={
            "description": "Allow http registry",
            "sample": False,
        },
        default=False,
    )
    allow_insecure_registry: bool = field(
        init=False,
        metadata={
            "description": "Allow insecure registry",
            "sample": False,
        },
        default=False,
    )
    rekor_url: str = field(
        init=False,
        metadata={
            "description": "URL for rekor stl server",
            "sample": "https://rekor.sigstore.dev",
        },
        default="https://rekor.sigstore.dev",
    )
    env_variables: Dict[str, str] = field(
        init=False,
        metadata={
            "description": "environment variables used for signing",
            "sample": '{"AWS_ACCESS_KEY_ID": "xxxxxxx", "AWS_SECRET_ACCESS_KEY":"yyyyyyyyy"}',
        },
        default_factory=dict,
    )
    upload_tlog: bool = field(
        init=False,
        metadata={"description": "upload signing record to rekor", "sample": "False"},
        default=True,
    )

    log_level: str = field(
        init=False, metadata={"description": "Log level", "sample": "debug"}, default="info"
    )
    key_aliases: Dict[str, str] = field(
        init=False,
        metadata={
            "description": "Aliases for signing keys",
            "sample": "{'production':'abcde1245'}",
        },
        default_factory=dict,
    )

    registry_user: str = field(
        init=False,
        metadata={"description": "Registry basic user", "sample": "username"},
        default="",
    )

    registry_password: str = field(
        init=False,
        metadata={"description": "Registry basic password", "sample": "password"},
        default="",
    )
    registry_auth_file: str = field(
        init=False,
        metadata={"description": "Registry basic auth file", "sample": "auth.json"},
        default="",
    )
    retries: int = field(
        init=False,
        metadata={"description": "Number of retries for http requests", "sample": 5},
        default=5,
    )

    SUPPORTED_OPERATIONS: ClassVar[List[Type[SignOperation]]] = [
        ContainerSignOperation,
    ]

    _signer_config_key: str = "cosign_signer"

    def __post_init__(self) -> None:
        """Post initialization of the class."""
        set_log_level(LOG, self.log_level)
        self.container_registry_client = ContainerRegistryClient(
            username=self.registry_user,
            password=self.registry_password,
            auth_file=self.registry_auth_file,
            log_level=self.log_level,
        )
        self.auth_token = AuthTokenWrapper(token="")

    def load_config(self: CosignSigner, config_data: Dict[str, Any]) -> None:
        """Load configuration of messaging signer.

        Arguments:
            config_data (dict): configuration data to load
        """
        self.cosign_bin = config_data["cosign_signer"].get("cosign_bin", self.cosign_bin)
        self.timeout = config_data["cosign_signer"].get("timeout", self.timeout)
        self.allow_http_registry = config_data["cosign_signer"].get(
            "allow_http_registry", self.allow_http_registry
        )
        self.allow_insecure_registry = config_data["cosign_signer"].get(
            "allow_insecure_registry", self.allow_insecure_registry
        )
        self.rekor_url = config_data["cosign_signer"].get("rekor_url", self.rekor_url)
        self.upload_tlog = config_data["cosign_signer"].get("upload_tlog", self.upload_tlog)
        self.env_variables = config_data["cosign_signer"].get("env_variables", self.env_variables)
        self.key_aliases = config_data["cosign_signer"].get("key_aliases", {})
        self.registry_user = config_data["cosign_signer"].get("registry_user", self.registry_user)
        self.registry_password = config_data["cosign_signer"].get(
            "registry_password", self.registry_password
        )
        self.retries = config_data["cosign_signer"].get("retries", self.retries)
        self.container_registry_client = ContainerRegistryClient(
            username=self.registry_user,
            password=self.registry_password,
            auth_file=self.registry_auth_file,
            log_level=self.log_level,
        )
        self.num_threads = config_data["cosign_signer"].get("num_threads", self.num_threads)

    def operations(self: CosignSigner) -> List[Type[SignOperation]]:
        """Return list of supported signing operation classes.

        Returns:
            List[Type[SignOperation]]: list of supported operations
        """
        return self.SUPPORTED_OPERATIONS

    def _sign_container(
        self,
        args: List[str],
        env: Dict[str, str],
        tries: int,
        identity: Optional[str] = None,
        ref_digest: Optional[str] = None,
        tag: Optional[str] = None,
    ) -> Any:
        LOG.info(f"Signing {identity} ({ref_digest}{tag or ''})")
        return run_command(args, env=env, tries=tries)

    def sign(self: CosignSigner, operation: SignOperation) -> SigningResults:
        """Run signing operation.

        Arguments:
            operation (SignOperation): signing operation to run

        Returns:
            SigningResults: results of the signing operation
        """
        if isinstance(operation, ContainerSignOperation):
            return self.container_sign(operation)
        else:
            raise UnsupportedOperation(operation)

    def container_sign(self: CosignSigner, operation: ContainerSignOperation) -> SigningResults:
        """Run container signing operation.

        Arguments:
            operation (ContainerSignOperation): container signing operation to run

        Returns:
            SigningResults: results of the container signing operation
        """
        if operation.references and len(operation.digests) != len(operation.references):
            raise ValueError("Digests must pair with references")

        signer_results = CosignSignerResults(status="ok", error_message="")

        operation_result = ContainerSignResult(
            signing_keys=operation.signing_keys, results=[], failed=False
        )
        signing_keys = operation.signing_keys

        signing_results = SigningResults(
            signer=self,
            operation=operation,
            signer_results=signer_results,
            operation_result=operation_result,
        )

        for _signing_key in signing_keys:
            if _signing_key in self.key_aliases:
                signing_key = self.key_aliases[_signing_key]
                LOG.info(f"Using signing key alias {signing_key} for {_signing_key}")
            else:
                signing_key = _signing_key

            ref_args_group: dict[str, List[Tuple[List[str], Dict[str, Any]]]] = {}
            common_args = [
                self.cosign_bin,
                "-t",
                self.timeout,
                "sign",
                "-y",
                "--key",
                signing_key,
                "--allow-http-registry=%s" % ("true" if self.allow_http_registry else "false"),
                "--allow-insecure-registry=%s"
                % ("true" if self.allow_insecure_registry else "false"),
                "--rekor-url",
                self.rekor_url,
                "--tlog-upload=%s" % ("true" if self.upload_tlog else "false"),
            ]
            if self.registry_user:
                common_args += ["--registry-username", self.registry_user]
            if self.registry_password:
                common_args += ["--registry-password", self.registry_password]
            env_vars = os.environ.copy()
            env_vars.update(self.env_variables)
            if operation.references:
                for ref, identity, digest in itertools.zip_longest(
                    operation.references,
                    operation.identity_references,
                    operation.digests,
                    fillvalue="",
                ):
                    args = []
                    named_args = {}
                    if identity:
                        args = ["--sign-container-identity", identity]
                        named_args["identity"] = identity
                    repo, tag = ref.rsplit(":", 1)
                    args.extend(["-a", f"tag={tag}", f"{repo}@{digest}"])
                    named_args["tag"] = tag
                    # To avoid conflict caused by running in parallel for the same reference, group
                    # args by reference.
                    ref_digest = f"{repo}@{digest}"
                    named_args["ref_digest"] = digest

                    ref_args_group.setdefault(ref_digest, [])
                    ref_args_group[ref_digest].append((args, named_args))

            else:
                for ref_digest, identity in itertools.zip_longest(
                    operation.digests, operation.identity_references, fillvalue=""
                ):
                    args = []
                    named_args = {}
                    if identity:
                        args = ["--sign-container-identity", identity]
                        named_args["identity"] = identity
                    named_args["ref_digest"] = ref_digest
                    args.append(ref_digest)
                    ref_args_group.setdefault(ref_digest, [])
                    ref_args_group[ref_digest].append((args, named_args))

            # Execute cosign commands serially in each group while running groups concurrently.
            ret = run_in_parallel(
                lambda args_group, **kwargs: [
                    self._sign_container(
                        common_args + args, env=env_vars, tries=self.retries, **kwargs
                    )
                    for args, kwargs in args_group
                ],
                [
                    FData(
                        args=[args_group],
                    )
                    for args_group in ref_args_group.values()
                ],
                self.num_threads,
            )

            for stdout, stderr, returncode in itertools.chain(*ret.values()):
                if returncode != 0:
                    operation_result.results.append(stderr)
                    operation_result.failed = True
                    signing_results.signer_results.status = "failed"
                    signing_results.signer_results.error_message += stderr
                else:
                    operation_result.results.append(stderr)
            signing_results.operation_result = operation_result
        return signing_results

    def existing_signatures(self, reference: str) -> Tuple[bool, str]:
        """Return list of existing signatures for given reference.

        Args:
            reference (str): reference to get list of signatures for

        Returns:
            Tuple[bool, str]: tuple of success flag and error message or result string
        """
        common_args = [
            self.cosign_bin,
            "-t",
            self.timeout,
            "triangulate",
            "--allow-http-registry=%s" % ("true" if self.allow_http_registry else "false"),
            "--allow-insecure-registry=%s" % ("true" if self.allow_insecure_registry else "false"),
        ]
        if self.registry_user:
            common_args += ["--registry-username", self.registry_user]
        if self.registry_password:
            common_args += ["--registry-password", self.registry_password]
        env_vars = os.environ.copy()
        env_vars.update(self.env_variables)
        stdout, stderr, returncode = run_command(
            common_args + [reference],
            env=env_vars,
        )
        if returncode != 0:
            return False, stderr
        else:
            ret, err_msg = self.container_registry_client.check_container_image_exists(
                stdout.strip("\n"), auth_token=self.auth_token
            )
            if ret:
                return True, stdout.split("\n")
            elif err_msg:
                return False, err_msg
            return True, ""


def cosign_container_sign(
    signing_keys: List[str] = [],
    config_file: str = "",
    digest: List[str] = [],
    reference: List[str] = [],
    identity: List[str] = [],
) -> Dict[str, Any]:
    """Run containersign operation with cli arguments.

    Args:
        signing_key (str): path to the signing key
        config_file (str): path to the config file
        digest (str): digest of the image to sign
        reference (str): reference of the image to sign
        identity (str): identity to sign the image with
    Returns:
        dict: signing result
    """
    cosign_signer = CosignSigner()
    config = _get_config_file(config_file)
    cosign_signer.load_config(load_config(os.path.expanduser(config)))

    operation = ContainerSignOperation(
        digests=digest,
        references=reference,
        identity_references=identity,
        signing_keys=signing_keys,
        task_id="",
    )
    signing_result = cosign_signer.sign(operation)
    return {
        "signer_result": signing_result.signer_results.to_dict(),
        "operation_results": signing_result.operation_result.results,
        "operation": signing_result.operation.to_dict(),
        "signing_keys": signing_result.operation_result.signing_keys,
    }


@click.command()
@click.option(
    "--signing-key",
    required=True,
    multiple=True,
    help="signing key used by cosign.",
)
@click.option("--config-file", default=CONFIG_PATHS[0], help="path to the config file")
@click.option(
    "--digest",
    required=True,
    multiple=True,
    type=str,
    help="Digests which should be signed.",
)
@click.option(
    "--reference",
    required=False,
    multiple=True,
    type=str,
    help="References which should be signed.",
)
@click.option(
    "--identity",
    required=False,
    multiple=True,
    type=str,
    help="Identity reference.",
)
@click.option("--raw", default=False, is_flag=True, help="Print raw output instead of json")
def cosign_container_sign_main(
    signing_key: List[str] = [],
    config_file: str = "",
    digest: List[str] = [],
    reference: List[str] = [],
    identity: List[str] = [],
    raw: bool = False,
) -> None:
    """Entry point method for containersign operation."""
    ret = cosign_container_sign(
        signing_keys=signing_key,
        config_file=config_file,
        digest=digest,
        reference=reference,
        identity=identity,
    )
    if not raw:
        click.echo(json.dumps(ret))
        if ret["signer_result"]["status"] == "error":
            sys.exit(1)
    else:
        if ret["signer_result"]["status"] == "error":
            print(ret["signer_result"]["error_message"], file=sys.stderr)
            sys.exit(1)
        else:
            for claim in ret["operation_results"]:
                print(claim)


def cosign_list_existing_signatures(config_file: str, reference: str) -> Tuple[bool, str]:
    """List existing signatures for given reference.

    Args:
        config_file (str): path to the config file
        reference (str): reference to get list of signatures for
    Returns:
        Tuple[bool, str]: tuple of success flag and error message or result string
    """
    cosign_signer = CosignSigner()
    config = _get_config_file(config_file)
    cosign_signer.load_config(load_config(os.path.expanduser(config)))
    return cosign_signer.existing_signatures(reference)



from __future__ import annotations
import dataclasses

from typing import List, ClassVar, Any, Dict
from typing_extensions import Self

from ..results.operation_result import OperationResult


@dataclasses.dataclass
class BlobSignResult(OperationResult):
    """BlobSignResult model.

    Attributes:
        results (List[Any]): List of signing result outputs.
        signing_keys (List[str]): The signing keys used during signing.
        failed (bool): Indicates if the operation failed.
    """

    ResultType: ClassVar[OperationResult]
    results: List[Any]
    signing_keys: List[str]
    failed: bool

    def to_dict(self: Self) -> Dict[Any, Any]:
        """Return dict representation of BlobSignResult."""
        return {"results": self.results, "signing_keys": self.signing_keys, "failed": self.failed}



from unittest.mock import patch
import tempfile

from pytest import fixture


@fixture
def f_cosign_signer(f_config_msg_signer_ok):
    with patch("pubtools.sign.signers.cosignsigner.CosignSigner") as signer:
        yield signer


@fixture
def f_config_cosign_signer_ok(f_client_certificate):
    with tempfile.NamedTemporaryFile() as tmpf:
        tmpf.write(
            """
cosign_signer:
  timeout: 30s
  rekor_url: https://rekor.sigstore.dev
  registry_user: some-user
  registry_password: some-password
  retries: 1
  log_level: debug
        """.encode(
                "utf-8"
            )
        )
        tmpf.flush()
        yield tmpf.name


@fixture
def f_config_cosign_signer_aliases(f_client_certificate):
    with tempfile.NamedTemporaryFile() as tmpf:
        tmpf.write(
            """
cosign_signer:
  timeout: 30s
  rekor_url: https://rekor.sigstore.dev
  log_level: debug
  key_aliases:
    beta: abcde1245
        """.encode(
                "utf-8"
            )
        )
        tmpf.flush()
        yield tmpf.name


@fixture
def f_config_cosign_signer_no_auth(f_client_certificate):
    with tempfile.NamedTemporaryFile() as tmpf:
        tmpf.write(
            """
cosign_signer:
  timeout: 30s
  rekor_url: https://rekor.sigstore.dev
  log_level: debug
        """.encode(
                "utf-8"
            )
        )
        tmpf.flush()
        yield tmpf.name



import base64
import json
import os
import tempfile

from click.testing import CliRunner
import pytest
from unittest.mock import patch, Mock, ANY


from pubtools.sign.models.msg import MsgError
from pubtools.sign.signers.msgsigner import (
    MsgSigner,
    MsgSignerResults,
    msg_clear_sign,
    ContainerSignOperation,
    ContainerSignResult,
    BlobSignResult,
    ClearSignOperation,
    BlobSignOperation,
    ClearSignResult,
    SignRequestType,
    msg_clear_sign_main,
    msg_container_sign_main,
    msg_blob_sign_main,
    _get_config_file,
)
from pubtools.sign.conf.conf import load_config
from pubtools.sign.models.msg import MsgMessage
from pubtools.sign.exceptions import UnsupportedOperation
from pubtools.sign.results.signing_results import SigningResults


def test_msg_container_sign(f_msg_signer, f_config_msg_signer_ok):
    f_msg_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "ok"
    }
    f_msg_signer.return_value.sign.return_value.operation_result.results = []
    f_msg_signer.return_value.sign.return_value.operation_result.signing_keys = []
    f_msg_signer.return_value.sign.return_value.operation.to_dict.return_value = {}
    result = CliRunner().invoke(
        msg_container_sign_main,
        [
            "--signing-key",
            "test-signing-key",
            "--digest",
            "some-digest",
            "--reference",
            "some-reference",
            "--task-id",
            "1",
            "--config-file",
            f_config_msg_signer_ok,
        ],
    )
    print(result.stdout)
    assert result.exit_code == 0, result.output


def test_msg_container_sign_keyname(f_msg_signer, f_config_msg_signer_ok):
    f_msg_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "ok"
    }
    f_msg_signer.return_value.sign.return_value.operation_result.results = []
    f_msg_signer.return_value.sign.return_value.operation_result.signing_keys = []
    f_msg_signer.return_value.sign.return_value.operation.to_dict.return_value = {}
    result = CliRunner().invoke(
        msg_container_sign_main,
        [
            "--signing-key",
            "test-signing-key",
            "--signing-key-name",
            "test-signing-key-name",
            "--digest",
            "some-digest",
            "--reference",
            "some-reference",
            "--task-id",
            "1",
            "--config-file",
            f_config_msg_signer_ok,
        ],
    )
    print(result.stdout)
    assert result.exit_code == 0, result.output


def test_msg_container_sign_requester(f_msg_signer, f_config_msg_signer_ok):
    f_msg_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "ok"
    }
    f_msg_signer.return_value.sign.return_value.operation_result.results = []
    f_msg_signer.return_value.sign.return_value.operation_result.signing_keys = []
    f_msg_signer.return_value.sign.return_value.operation.to_dict.return_value = {}
    result = CliRunner().invoke(
        msg_container_sign_main,
        [
            "--signing-key",
            "test-signing-key",
            "--digest",
            "some-digest",
            "--reference",
            "some-reference",
            "--task-id",
            "1",
            "--config-file",
            f_config_msg_signer_ok,
            "--requester",
            "test-requester",
        ],
    )
    print(result.stdout)
    assert result.exit_code == 0, result.output


def test_msg_container_sign_error(f_msg_signer, f_config_msg_signer_ok):
    f_msg_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "error",
        "error_message": "simulated error",
    }
    f_msg_signer.return_value.sign.return_value.operation_result.results = []
    f_msg_signer.return_value.sign.return_value.operation_result.signing_keys = []
    f_msg_signer.return_value.sign.return_value.operation.to_dict.return_value = {}
    result = CliRunner().invoke(
        msg_container_sign_main,
        [
            "--signing-key",
            "test-signing-key",
            "--digest",
            "some-digest",
            "--reference",
            "some-reference",
            "--task-id",
            "1",
            "--config-file",
            f_config_msg_signer_ok,
        ],
    )
    print(result.stdout)
    assert result.exit_code == 1, result.output


def test_msg_container_sign_raw(f_msg_signer, f_config_msg_signer_ok):
    f_msg_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "ok"
    }
    f_msg_signer.return_value.sign.return_value.operation_result.results = [
        ({"i": 456, "msg": {"errors": [], "signed_claim": "signed"}}, {})
    ]
    f_msg_signer.return_value.sign.return_value.operation_result.signing_keys = []
    result = CliRunner().invoke(
        msg_container_sign_main,
        [
            "--signing-key",
            "test-signing-key",
            "--digest",
            "some-digest",
            "--reference",
            "some-reference",
            "--task-id",
            "1",
            "--config-file",
            f_config_msg_signer_ok,
            "--raw",
        ],
    )
    print(result.stdout)
    assert result.exit_code == 0, result.output
    assert result.output == "signed\n"


def test_msg_container_sign_raw_error(f_msg_signer, f_config_msg_signer_ok):
    f_msg_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "ok",
        "error_message": "",
    }
    f_msg_signer.return_value.sign.return_value.operation_result.results = [
        (
            {
                "i": 456,
                "msg": {"errors": ['no signing key "test-signing-key"'], "signed_claim": None},
            },
            {},
        )
    ]
    f_msg_signer.return_value.sign.return_value.operation_result.signing_keys = []
    result = CliRunner().invoke(
        msg_container_sign_main,
        [
            "--signing-key",
            "test-signing-key",
            "--digest",
            "some-digest",
            "--reference",
            "some-reference",
            "--task-id",
            "1",
            "--config-file",
            f_config_msg_signer_ok,
            "--raw",
        ],
    )
    print(result.stdout)
    assert result.exit_code == 1, result.output
    assert result.output == 'no signing key "test-signing-key"\n'


def test_msg_clearsign_sign(f_msg_signer, f_config_msg_signer_ok):
    f_msg_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "ok"
    }
    f_msg_signer.return_value.sign.return_value.operation_result.outputs = [
        ({"i": 456, "msg": {"errors": [], "signed_data": "test"}}, {})
    ]
    f_msg_signer.return_value.sign.return_value.operation_result.signing_keys = []
    f_msg_signer.return_value.sign.return_value.operation.to_dict.return_value = {}
    result = CliRunner().invoke(
        msg_clear_sign_main,
        [
            "--signing-key",
            "test-signing-key",
            "--task-id",
            "1",
            "--config-file",
            f_config_msg_signer_ok,
            "hello world",
        ],
    )
    print(result.output)
    assert result.exit_code == 0, result


def test_msg_clearsign_sign_requester(f_msg_signer, f_config_msg_signer_ok):
    f_msg_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "ok"
    }
    f_msg_signer.return_value.sign.return_value.operation_result.outputs = [
        ({"i": 456, "msg": {"errors": [], "signed_data": "test"}}, {})
    ]
    f_msg_signer.return_value.sign.return_value.operation_result.signing_keys = []
    f_msg_signer.return_value.sign.return_value.operation.to_dict.return_value = {}
    result = CliRunner().invoke(
        msg_clear_sign_main,
        [
            "--signing-key",
            "test-signing-key",
            "--task-id",
            "1",
            "--requester",
            "test-requester",
            "--config-file",
            f_config_msg_signer_ok,
            "hello world",
        ],
    )
    print(result.output)
    assert result.exit_code == 0, result


def test_msg_clearsign_sign_error(f_msg_signer, f_config_msg_signer_ok):
    f_msg_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "error",
        "error_message": "simulated error",
    }
    f_msg_signer.return_value.sign.return_value.operation_result.outputs = []
    f_msg_signer.return_value.sign.return_value.operation_result.signing_keys = []
    f_msg_signer.return_value.sign.return_value.operation.to_dict.return_value = {}
    result = CliRunner().invoke(
        msg_clear_sign_main,
        [
            "--signing-key",
            "test-signing-key",
            "--task-id",
            "1",
            "--config-file",
            f_config_msg_signer_ok,
            "hello world",
        ],
    )
    assert result.exit_code == 1, result.output


def test_msg_clearsign_sign_raw(f_msg_signer, f_config_msg_signer_ok):
    f_msg_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "ok"
    }
    f_msg_signer.return_value.sign.return_value.operation_result.outputs = [
        ({"i": 456, "msg": {"errors": [], "signed_data": "signed"}}, {})
    ]
    f_msg_signer.return_value.sign.return_value.operation_result.signing_keys = []
    result = CliRunner().invoke(
        msg_clear_sign_main,
        [
            "--signing-key",
            "test-signing-key",
            "--task-id",
            "1",
            "--config-file",
            f_config_msg_signer_ok,
            "--raw",
            "hello world",
        ],
    )
    assert result.exit_code == 0, result.output
    assert result.output == "signed\n"


def test_msg_clearsign_sign_raw_error2(f_msg_signer, f_config_msg_signer_ok):
    f_msg_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "error",
        "error_message": "simulated error",
    }
    f_msg_signer.return_value.sign.return_value.operation_result.outputs = []
    f_msg_signer.return_value.sign.return_value.operation_result.signing_keys = []
    result = CliRunner().invoke(
        msg_clear_sign_main,
        [
            "--signing-key",
            "test-signing-key",
            "--task-id",
            "1",
            "--config-file",
            f_config_msg_signer_ok,
            "--raw",
            "hello world",
        ],
    )
    print(result.stdout)
    assert result.exit_code == 1, result.output
    assert result.output == "simulated error\n"


def test_msg_clearsign_sign_raw_error(f_msg_signer, f_config_msg_signer_ok):
    f_msg_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "ok",
    }
    f_msg_signer.return_value.sign.return_value.operation_result.outputs = [
        ({"i": 456, "msg": {"errors": ["simulated error"], "signed_data": "signed"}}, {})
    ]
    f_msg_signer.return_value.sign.return_value.operation_result.signing_keys = []
    result = CliRunner().invoke(
        msg_clear_sign_main,
        [
            "--signing-key",
            "test-signing-key",
            "--task-id",
            "1",
            "--config-file",
            f_config_msg_signer_ok,
            "--raw",
            "hello world",
        ],
    )
    print(result.stdout)
    assert result.exit_code == 1, result.output
    assert result.output == "simulated error\n"


def test_msg_clearsign_sign_file_input(f_msg_signer, f_config_msg_signer_ok):
    f_msg_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "ok"
    }
    f_msg_signer.return_value.sign.return_value.operation_result.outputs = []
    f_msg_signer.return_value.sign.return_value.operation_result.signing_keys = []
    f_msg_signer.return_value.sign.return_value.operation.to_dict.return_value = {}
    result = CliRunner().invoke(
        msg_clear_sign_main,
        [
            "--signing-key",
            "test-signing-key",
            "--task-id",
            "1",
            "--config-file",
            f_config_msg_signer_ok,
            f"@{f_config_msg_signer_ok}",
        ],
    )
    assert result.exit_code == 0, result.output


def test__msg_clearsign_sign(f_msg_signer, f_config_msg_signer_ok):
    msg_clear_sign(
        ["hello world"],
        signing_keys=["test-signing-key"],
        task_id="1",
        config_file=f_config_msg_signer_ok,
        repo="repo",
    )

    f_msg_signer.return_value.load_config.assert_called_with(load_config(f_config_msg_signer_ok))
    operation = ClearSignOperation(
        inputs=["hello world"], signing_keys=["test-signing-key"], task_id="1", repo="repo"
    )
    f_msg_signer.return_value.sign.assert_called_with(operation)


def test__msg_clearsign_sign_file_input(f_msg_signer, f_config_msg_signer_ok):
    msg_clear_sign(
        [f"@{f_config_msg_signer_ok}"],
        signing_keys=["test-signing-key"],
        task_id="1",
        config_file=f_config_msg_signer_ok,
        repo="repo",
    )

    f_msg_signer.return_value.load_config.assert_called_with(load_config(f_config_msg_signer_ok))
    operation = ClearSignOperation(
        inputs=[open(f_config_msg_signer_ok).read()],
        signing_keys=["test-signing-key"],
        task_id="1",
        repo="repo",
    )
    f_msg_signer.return_value.sign.assert_called_with(operation)


def test_get_config_file(f_config_msg_signer_ok):
    assert _get_config_file(f_config_msg_signer_ok) == f_config_msg_signer_ok


def test_get_config_file_default(f_config_msg_signer_ok):
    def _patched_exist(fname):
        if fname == "/etc/pubtools-sign/conf.yaml":
            return True
        else:
            return False

    with patch("os.path.exists", side_effect=_patched_exist):
        assert _get_config_file("/non-existining/file") == "/etc/pubtools-sign/conf.yaml"


def test_get_config_no_configuration_found():
    with patch("os.path.exists", return_value=False):
        with pytest.raises(ValueError):
            _get_config_file("/non-existining/file")


def test__construct_signing_message(f_config_msg_signer_ok):
    signer = MsgSigner()
    signer.load_config(load_config(f_config_msg_signer_ok))
    with patch("uuid.uuid4", return_value="1234-5678-abcd-efgh"):
        with patch("pubtools.sign.signers.msgsigner.isodate_now") as patched_date:
            patched_date.return_value = "created-date-Z"
            ret = signer._construct_signing_message(
                "some-claim", "some-signing-key", "repo", extra_attrs={}, signing_key_name=""
            )
            assert ret == {
                "sig_key_id": "ning-key",
                "claim_file": "some-claim",
                "request_id": "1234-5678-abcd-efgh",
                "created": "created-date-Z",
                "requested_by": "pubtools-sign-test",
                "repo": "repo",
            }


def test__construct_signing_message_signing_key_name(f_config_msg_signer_ok):
    signer = MsgSigner()
    signer.load_config(load_config(f_config_msg_signer_ok))
    with patch("uuid.uuid4", return_value="1234-5678-abcd-efgh"):
        with patch("pubtools.sign.signers.msgsigner.isodate_now") as patched_date:
            patched_date.return_value = "created-date-Z"
            ret = signer._construct_signing_message(
                "some-claim",
                "some-signing-key",
                "repo",
                extra_attrs={},
                signing_key_name="test-signing-key-name",
            )
            assert ret == {
                "sig_key_id": "ning-key",
                "sig_keyname": "test-signing-key-name",
                "claim_file": "some-claim",
                "request_id": "1234-5678-abcd-efgh",
                "created": "created-date-Z",
                "requested_by": "pubtools-sign-test",
                "repo": "repo",
            }


def test__construct_headers(f_config_msg_signer_ok):
    signer = MsgSigner()
    signer.load_config(load_config(f_config_msg_signer_ok))
    with patch("uuid.uuid4", return_value="1234-5678-abcd-efgh"):
        with patch("pubtools.sign.signers.msgsigner.isodate_now") as patched_date:
            patched_date.return_value = "created-date-Z"
            ret = signer._construct_headers(
                SignRequestType.CONTAINER, extra_attrs={"extra": "extra"}
            )
            assert ret == {
                "service": "pubtools-sign",
                "environment": "prod",
                "owner_id": "pubtools-sign-test",
                "mtype": SignRequestType.CONTAINER,
                "source": "metadata",
                "extra": "extra",
            }


def test_operations():
    signer = MsgSigner()
    assert signer.operations() == [ContainerSignOperation, ClearSignOperation, BlobSignOperation]


def test_create_msg_message(f_config_msg_signer_ok):
    signer = MsgSigner()
    signer.load_config(load_config(f_config_msg_signer_ok))

    data = "test-data"
    with patch("uuid.uuid4", return_value="1234-5678-abcd-efgh"):
        with patch("pubtools.sign.signers.msgsigner.isodate_now") as patched_date:
            patched_date.return_value = "created-date-Z"
            operation = ClearSignOperation(
                inputs=["test-data-inputs"], signing_keys=["test-key"], task_id="1", repo="repo"
            )
            created_messages = signer._create_msg_messages(
                data, "repo", operation, SignRequestType.CONTAINER
            )
            assert len(created_messages) == 1
            assert created_messages == [
                MsgMessage(
                    headers={
                        "service": "pubtools-sign",
                        "environment": "prod",
                        "owner_id": "pubtools-sign-test",
                        "mtype": SignRequestType.CONTAINER,
                        "source": "metadata",
                    },
                    address="topic://Topic.sign",
                    body={
                        "sig_key_id": "test-key",
                        "claim_file": "test-data",
                        "request_id": "1234-5678-abcd-efgh",
                        "created": "created-date-Z",
                        "requested_by": "pubtools-sign-test",
                        "repo": "repo",
                    },
                )
            ]
            assert signer._create_msg_messages(
                data, "repo", operation, SignRequestType.CLEARSIGN
            ) == [
                MsgMessage(
                    headers={
                        "service": "pubtools-sign",
                        "environment": "prod",
                        "owner_id": "pubtools-sign-test",
                        "mtype": SignRequestType.CLEARSIGN,
                        "source": "metadata",
                    },
                    address="topic://Topic.sign",
                    body={
                        "sig_key_id": "test-key",
                        "data": "test-data",
                        "request_id": "1234-5678-abcd-efgh",
                        "created": "created-date-Z",
                        "requested_by": "pubtools-sign-test",
                        "repo": "repo",
                    },
                )
            ]


def test_sign(f_config_msg_signer_ok):
    signer = MsgSigner()
    signer.load_config(load_config(f_config_msg_signer_ok))
    container_sign_operation = ContainerSignOperation(
        digests=("some-digest",),
        references=("some-reference",),
        signing_keys=["test-signing-key"],
        task_id="1",
    )
    clear_sign_operation = ClearSignOperation(
        inputs=["hello world"], signing_keys=["test-signing-key"], task_id="1", repo="repo"
    )

    with patch("pubtools.sign.signers.msgsigner.MsgSigner.clear_sign") as patched_clear_sign:
        with patch(
            "pubtools.sign.signers.msgsigner.MsgSigner.container_sign"
        ) as patched_container_sign:
            signer.sign(container_sign_operation)
            patched_container_sign.assert_called_once()
            signer.sign(clear_sign_operation)
            patched_clear_sign.assert_called_once()
            with pytest.raises(UnsupportedOperation):
                signer.sign(Mock())


def test_create_manifest_claim_message():
    signer = MsgSigner()
    assert signer.create_manifest_claim_message(
        "some-digest", "some-reference"
    ) == base64.b64encode(
        json.dumps(
            {
                "critical": {
                    "type": "atomic container signature",
                    "image": {"docker-manifest-digest": "some-digest"},
                    "identity": {"docker-reference": "some-reference"},
                },
                "optional": {"creator": "pubtools-sign"},
            }
        ).encode("latin-1")
    ).decode(
        "latin-1"
    )


@patch("uuid.uuid4", return_value="1234-5678-abcd-efgh")
def test_clear_sign(patched_uuid, f_config_msg_signer_ok):
    clear_sign_operation = ClearSignOperation(
        inputs=["hello world"], signing_keys=["test-signing-key"], task_id="1", repo="repo"
    )
    with patch("pubtools.sign.signers.msgsigner.SendClient") as patched_send_client:
        with patch("pubtools.sign.signers.msgsigner.RecvClient") as patched_recv_client:
            patched_send_client.return_value.run.return_value = []
            patched_recv_client.return_value.run.return_value = []
            patched_recv_client.return_value.recv = {"1234-5678-abcd-efgh": "signed:'hello world'"}
            patched_recv_client.return_value._errors = []
            patched_recv_client.return_value.get_errors.return_value = []

            signer = MsgSigner()
            signer.load_config(load_config(f_config_msg_signer_ok))
            res = signer.clear_sign(clear_sign_operation)

            assert res == SigningResults(
                signer=signer,
                operation=clear_sign_operation,
                signer_results=MsgSignerResults(status="ok", error_message=""),
                operation_result=ClearSignResult(
                    outputs=["signed:'hello world'"],
                    signing_keys=["test-signing-key"],
                ),
            )


@patch("uuid.uuid4", side_effect=["1234-5678-abcd-efgh", "2234-5678-abcd-efgh"])
def test_clear_sign_multiple_signing_keys(patched_uuid, f_config_msg_signer_ok):
    clear_sign_operation = ClearSignOperation(
        inputs=["hello world"],
        signing_keys=["test-signing-key", "test-signing-key2"],
        task_id="1",
        repo="repo",
    )
    with patch("pubtools.sign.signers.msgsigner.SendClient") as patched_send_client:
        with patch("pubtools.sign.signers.msgsigner.RecvClient") as patched_recv_client:
            patched_send_client.return_value.run.return_value = []
            patched_recv_client.return_value.run.return_value = []
            patched_recv_client.return_value.recv = {
                "1234-5678-abcd-efgh": "signed:'hello world'",
                "2234-5678-abcd-efgh": "signed:'hello world2'",
            }
            patched_recv_client.return_value.get_errors.return_value = []

            signer = MsgSigner()
            signer.load_config(load_config(f_config_msg_signer_ok))
            res = signer.clear_sign(clear_sign_operation)

            assert res == SigningResults(
                signer=signer,
                operation=clear_sign_operation,
                signer_results=MsgSignerResults(status="ok", error_message=""),
                operation_result=ClearSignResult(
                    outputs=["signed:'hello world'", "signed:'hello world2'"],
                    signing_keys=["test-signing-key", "test-signing-key2"],
                ),
            )


@patch("uuid.uuid4", return_value="1234-5678-abcd-efgh")
def test_clear_sign_aliases(patched_uuid, f_config_msg_signer_aliases):
    clear_sign_operation = ClearSignOperation(
        inputs=["hello world"], signing_keys=["beta"], task_id="1", repo="repo"
    )
    with patch("pubtools.sign.signers.msgsigner.SendClient") as patched_send_client:
        with patch("pubtools.sign.signers.msgsigner.RecvClient") as patched_recv_client:
            patched_send_client.return_value.run.return_value = []
            patched_recv_client.return_value.run.return_value = []
            patched_recv_client.return_value.recv = {"1234-5678-abcd-efgh": "signed:'hello world'"}
            patched_recv_client.return_value.get_errors.return_value = []

            signer = MsgSigner()
            signer.load_config(load_config(f_config_msg_signer_aliases))
            with patch(
                "pubtools.sign.signers.msgsigner.MsgSigner._construct_signing_message"
            ) as patch_construct_signing_message:
                patch_construct_signing_message.return_value = {
                    "request_id": "1234-5678-abcd-efgh",
                }
                res = signer.clear_sign(clear_sign_operation)
                patch_construct_signing_message.assert_called_once_with(
                    "aGVsbG8gd29ybGQ=",
                    "abcde1245",
                    "repo",
                    extra_attrs={"pub_task_id": "1"},
                    sig_type="clearsign_signature",
                    signing_key_name="",
                )

            assert res == SigningResults(
                signer=signer,
                operation=clear_sign_operation,
                signer_results=MsgSignerResults(status="ok", error_message=""),
                operation_result=ClearSignResult(
                    outputs=["signed:'hello world'"],
                    signing_keys=["beta"],
                ),
            )


@patch("uuid.uuid4", return_value="1234-5678-abcd-efgh")
def test_clear_sign_recv_errors(patched_uuid, f_config_msg_signer_ok):
    clear_sign_operation = ClearSignOperation(
        inputs=["hello world"], signing_keys=["test-signing-key"], task_id="1", repo="repo"
    )
    with patch("pubtools.sign.signers.msgsigner.SendClient") as patched_send_client:
        with patch("pubtools.sign.signers.msgsigner.RecvClient") as patched_recv_client:
            patched_send_client.return_value.run.return_value = []
            errors = [
                MsgError(
                    name="TestError", description="test error description", source="test-source"
                )
            ]
            patched_recv_client.return_value.get_errors.return_value = errors
            patched_recv_client.return_value._errors = errors
            patched_recv_client.return_value.recv = {"1234-5678-abcd-efgh": "signed:'hello world'"}

            signer = MsgSigner()
            signer.load_config(load_config(f_config_msg_signer_ok))
            res = signer.clear_sign(clear_sign_operation)

            assert res == SigningResults(
                signer=signer,
                operation=clear_sign_operation,
                signer_results=MsgSignerResults(
                    status="error", error_message="TestError : test error description\n"
                ),
                operation_result=ClearSignResult(outputs=[""], signing_keys=["test-signing-key"]),
            )


@patch("uuid.uuid4", return_value="1234-5678-abcd-efgh")
def test_clear_sign_send_errors(patched_uuid, f_config_msg_signer_ok):
    clear_sign_operation = ClearSignOperation(
        inputs=["hello world"], signing_keys=["test-signing-key"], task_id="1", repo="repo"
    )
    with patch("pubtools.sign.signers.msgsigner.SendClient") as patched_send_client:
        with patch("pubtools.sign.signers.msgsigner.RecvClient") as patched_recv_client:
            patched_send_client.return_value.run.return_value = [
                MsgError(
                    name="TestError", description="test error description", source="test-source"
                )
            ]
            patched_recv_client.return_value.run.return_value = []
            patched_recv_client.return_value.recv = {"1234-5678-abcd-efgh": "signed:'hello world'"}

            signer = MsgSigner()
            signer.load_config(load_config(f_config_msg_signer_ok))
            res = signer.clear_sign(clear_sign_operation)

            assert res == SigningResults(
                signer=signer,
                operation=clear_sign_operation,
                signer_results=MsgSignerResults(
                    status="error", error_message="TestError : test error description\n"
                ),
                operation_result=ClearSignResult(outputs=[""], signing_keys=["test-signing-key"]),
            )


@patch("uuid.uuid4", return_value="1234-5678-abcd-efgh")
def test_container_sign(
    patched_uuid, f_config_msg_signer_ok, f_client_certificate, f_ca_certificate
):
    container_sign_operation = ContainerSignOperation(
        task_id="1",
        digests=["sha256:abcdefg"],
        references=["some-registry/namespace/repo:tag"],
        signing_keys=["test-signing-key"],
    )

    with patch("pubtools.sign.signers.msgsigner.SendClient") as patched_send_client:
        with patch("pubtools.sign.signers.msgsigner.RecvClient") as patched_recv_client:
            patched_send_client.return_value.run.return_value = []
            patched_recv_client.return_value.run.return_value = []
            patched_recv_client.return_value.recv = {
                "1234-5678-abcd-efgh": (
                    {"msg": {"errors": [], "signed_claim": "signed:'claim'"}},
                    {"fake": "headers"},
                )
            }
            patched_recv_client.return_value.get_errors.return_value = []

            signer = MsgSigner()
            signer.load_config(load_config(f_config_msg_signer_ok))
            res = signer.container_sign(container_sign_operation)

            patched_send_client.assert_called_with(
                messages=[ANY],
                broker_urls=["amqps://broker-01:5671", "amqps://broker-02:5671"],
                cert=f_client_certificate,
                ca_cert=f_ca_certificate,
                retries=3,
                errors=patched_recv_client.return_value.get_errors.return_value,
            )

            assert res == SigningResults(
                signer=signer,
                operation=container_sign_operation,
                signer_results=MsgSignerResults(status="ok", error_message=""),
                operation_result=ContainerSignResult(
                    results=[
                        (
                            {"msg": {"errors": [], "signed_claim": "signed:'claim'"}},
                            {"fake": "headers"},
                        )
                    ],
                    signing_keys=["test-signing-key"],
                    failed=False,
                ),
            )


@patch("uuid.uuid4", side_effect=["1234-5678-abcd-efgh", "2234-5678-abcd-efgh"])
def test_container_sign_multiple_signing_keys(
    patched_uuid, f_config_msg_signer_ok, f_client_certificate, f_ca_certificate
):
    container_sign_operation = ContainerSignOperation(
        task_id="1",
        digests=["sha256:abcdefg"],
        references=["some-registry/namespace/repo:tag"],
        signing_keys=["test-signing-key", "test-signing-key2"],
    )

    with patch("pubtools.sign.signers.msgsigner.SendClient") as patched_send_client:
        with patch("pubtools.sign.signers.msgsigner.RecvClient") as patched_recv_client:
            patched_send_client.return_value.run.return_value = []
            patched_recv_client.return_value.run.return_value = []
            patched_recv_client.return_value.recv = {
                "1234-5678-abcd-efgh": (
                    {"msg": {"errors": [], "signed_claim": "signed:'claim1'"}},
                    {"fake": "headers"},
                ),
                "2234-5678-abcd-efgh": (
                    {"msg": {"errors": [], "signed_claim": "signed:'claim2'"}},
                    {"fake": "headers"},
                ),
            }
            patched_recv_client.return_value.get_errors.return_value = []

            signer = MsgSigner()
            signer.load_config(load_config(f_config_msg_signer_ok))
            res = signer.container_sign(container_sign_operation)

            patched_send_client.assert_called_with(
                messages=[ANY, ANY],
                broker_urls=["amqps://broker-01:5671", "amqps://broker-02:5671"],
                cert=f_client_certificate,
                ca_cert=f_ca_certificate,
                retries=3,
                errors=patched_recv_client.return_value.get_errors.return_value,
            )

            assert res == SigningResults(
                signer=signer,
                operation=container_sign_operation,
                signer_results=MsgSignerResults(status="ok", error_message=""),
                operation_result=ContainerSignResult(
                    results=[
                        (
                            {"msg": {"errors": [], "signed_claim": "signed:'claim1'"}},
                            {"fake": "headers"},
                        ),
                        (
                            {"msg": {"errors": [], "signed_claim": "signed:'claim2'"}},
                            {"fake": "headers"},
                        ),
                    ],
                    signing_keys=["test-signing-key", "test-signing-key2"],
                    failed=False,
                ),
            )


@patch("uuid.uuid4", return_value="1234-5678-abcd-efgh")
def test_container_sign_alias(patched_uuid, f_config_msg_signer_aliases, f_client_certificate):
    container_sign_operation = ContainerSignOperation(
        task_id="1",
        digests=["sha256:abcdefg"],
        references=["some-registry/namespace/repo:tag"],
        signing_keys=["beta"],
    )

    with patch("pubtools.sign.signers.msgsigner.SendClient") as patched_send_client:
        with patch("pubtools.sign.signers.msgsigner.RecvClient") as patched_recv_client:
            patched_send_client.return_value.run.return_value = []
            patched_recv_client.return_value.run.return_value = []
            patched_recv_client.return_value.recv = {
                "1234-5678-abcd-efgh": (
                    {"msg": {"errors": [], "signed_claim": "signed:'claim'"}},
                    {"fake": "headers"},
                )
            }
            patched_recv_client.return_value.get_errors.return_value = []

            signer = MsgSigner()
            signer.load_config(load_config(f_config_msg_signer_aliases))

            with patch(
                "pubtools.sign.signers.msgsigner.MsgSigner._construct_signing_message"
            ) as patch_construct_signing_message:
                patch_construct_signing_message.return_value = {
                    "request_id": "1234-5678-abcd-efgh",
                }
                res = signer.container_sign(container_sign_operation)
                patch_construct_signing_message.assert_called_once_with(
                    ANY,
                    "abcde1245",
                    "namespace/repo",
                    extra_attrs={"pub_task_id": "1", "manifest_digest": "sha256:abcdefg"},
                    signing_key_name="",
                    sig_type="container_signature",
                )

            patched_send_client.assert_called_with(
                messages=[ANY],
                broker_urls=["amqps://broker-01:5671", "amqps://broker-02:5671"],
                cert=f_client_certificate,
                ca_cert=os.path.expanduser("~/messaging/ca-cert.crt"),
                retries=3,
                errors=patched_recv_client.return_value.get_errors.return_value,
            )

            assert res == SigningResults(
                signer=signer,
                operation=container_sign_operation,
                signer_results=MsgSignerResults(status="ok", error_message=""),
                operation_result=ContainerSignResult(
                    results=[
                        (
                            {"msg": {"errors": [], "signed_claim": "signed:'claim'"}},
                            {"fake": "headers"},
                        )
                    ],
                    signing_keys=["beta"],
                    failed=False,
                ),
            )


@patch("uuid.uuid4", return_value="1234-5678-abcd-efgh")
def test_container_sign_recv_errors(patched_uuid, f_config_msg_signer_ok):
    container_sign_operation = ContainerSignOperation(
        task_id="1",
        digests=["sha256:abcdefg"],
        references=["some-registry/namespace/repo:tag"],
        signing_keys=["test-signing-key"],
    )

    with patch("pubtools.sign.signers.msgsigner.SendClient") as patched_send_client:
        with patch("pubtools.sign.signers.msgsigner.RecvClient") as patched_recv_client:
            patched_send_client.return_value.run.return_value = []
            errors = [
                MsgError(
                    name="TestError", description="test error description", source="test-source"
                )
            ]
            patched_recv_client.return_value._errors = errors
            patched_recv_client.return_value.get_errors.return_value = errors
            patched_recv_client.return_value.recv = {"1234-5678-abcd-efgh": "signed:'hello world'"}

            signer = MsgSigner()
            signer.load_config(load_config(f_config_msg_signer_ok))
            res = signer.container_sign(container_sign_operation)

            assert res == SigningResults(
                signer=signer,
                operation=container_sign_operation,
                signer_results=MsgSignerResults(
                    status="error", error_message="TestError : test error description\n"
                ),
                operation_result=ContainerSignResult(
                    results=[""], signing_keys=["test-signing-key"], failed=False
                ),
            )


@patch("uuid.uuid4", return_value="1234-5678-abcd-efgh")
def test_container_sign_send_errors(patched_uuid, f_config_msg_signer_ok):
    container_sign_operation = ContainerSignOperation(
        task_id="1",
        digests=["sha256:abcdefg"],
        references=["some-registry/namespace/repo:tag"],
        signing_keys=["test-signing-key"],
    )
    with patch("pubtools.sign.signers.msgsigner.SendClient") as patched_send_client:
        with patch("pubtools.sign.signers.msgsigner.RecvClient") as patched_recv_client:
            patched_send_client.return_value.run.return_value = [
                MsgError(
                    name="TestError", description="test error description", source="test-source"
                )
            ]
            patched_recv_client.return_value.run.return_value = []
            patched_recv_client.return_value.recv = {"1234-5678-abcd-efgh": "signed:'hello world'"}

            signer = MsgSigner()
            signer.load_config(load_config(f_config_msg_signer_ok))
            res = signer.container_sign(container_sign_operation)

            assert res == SigningResults(
                signer=signer,
                operation=container_sign_operation,
                signer_results=MsgSignerResults(
                    status="error", error_message="TestError : test error description\n"
                ),
                operation_result=ContainerSignResult(
                    results=[""], signing_keys=["test-signing-key"], failed=False
                ),
            )


@patch("uuid.uuid4", return_value="1234-5678-abcd-efgh")
def test_container_sign_wrong_inputs(patched_uuid, f_config_msg_signer_ok):
    container_sign_operation = ContainerSignOperation(
        task_id="1",
        digests=["sha256:abcdefg"],
        references=["some-registry/namespace/repo:tag", "some-registry/namespace/repo:tag2"],
        signing_keys=["test-signing-key"],
    )

    signer = MsgSigner()
    signer.load_config(load_config(f_config_msg_signer_ok))
    with pytest.raises(ValueError):
        signer.container_sign(container_sign_operation)


@patch("uuid.uuid4", return_value="1234-5678-abcd-efgh")
def test_container_sign_recv_timeout_fails(patched_uuid, f_config_msg_signer_ok):
    container_sign_operation = ContainerSignOperation(
        task_id="1",
        digests=["sha256:abcdefg"],
        references=["some-registry/namespace/repo:tag"],
        signing_keys=["test-signing-key"],
    )

    with patch("pubtools.sign.signers.msgsigner.SendClient") as patched_send_client:
        with patch("pubtools.sign.signers.msgsigner.RecvClient") as patched_recv_client:
            patched_send_client.return_value.run.return_value = []
            errors = [
                MsgError(
                    name="MessagingTimeout",
                    description="Out of time when receiving messages",
                    source="test-source",
                ),
                MsgError(
                    name="MessagingTimeout",
                    description="Out of time when receiving messages",
                    source="test-source",
                ),
                MsgError(
                    name="MessagingTimeout",
                    description="Out of time when receiving messages",
                    source="test-source",
                ),
            ]
            patched_recv_client.return_value.get_errors.return_value = errors
            patched_recv_client.return_value.recv = {
                "1234-5678-abcd-efgh": (
                    {"msg": {"errors": [], "signed_claim": "signed:'claim'"}},
                    {"fake": "headers"},
                )
            }

            signer = MsgSigner()
            signer.load_config(load_config(f_config_msg_signer_ok))
            signer.retries = 2
            signer.send_retries = 1

            res = signer.container_sign(container_sign_operation)

            assert res == SigningResults(
                signer=signer,
                operation=container_sign_operation,
                signer_results=MsgSignerResults(
                    status="error",
                    error_message="MessagingTimeout : Out of time when receiving messages\n"
                    "MessagingTimeout : Out of time when receiving messages\n",
                ),
                operation_result=ContainerSignResult(
                    results=[""], signing_keys=["test-signing-key"], failed=False
                ),
            )


@patch("uuid.uuid4", return_value="1234-5678-abcd-efgh")
def test_container_sign_recv_timeout_ok(patched_uuid, f_config_msg_signer_ok):
    container_sign_operation = ContainerSignOperation(
        task_id="1",
        digests=["sha256:abcdefg"],
        references=["some-registry/namespace/repo:tag"],
        signing_keys=["test-signing-key"],
    )

    with patch("pubtools.sign.signers.msgsigner.SendClient") as patched_send_client:
        with patch("pubtools.sign.signers.msgsigner.RecvClient") as patched_recv_client:
            patched_send_client.return_value.run.return_value = []
            errors = [
                MsgError(
                    name="MessagingTimeout",
                    description="Out of time when receiving messages",
                    source="test-source",
                ),
            ]
            patched_recv_client.return_value.get_errors.side_effect = lambda: errors

            patched_recv_client.return_value.recv = {
                "1234-5678-abcd-efgh": (
                    {"msg": {"errors": [], "signed_claim": "signed:'claim'"}},
                    {"fake": "headers"},
                )
            }

            signer = MsgSigner()

            signer.load_config(load_config(f_config_msg_signer_ok))
            signer.retries = 2
            signer.send_retries = 1

            res = signer.container_sign(container_sign_operation)

            assert res == SigningResults(
                signer=signer,
                operation=container_sign_operation,
                signer_results=MsgSignerResults(
                    status="ok",
                    error_message="",
                ),
                operation_result=ContainerSignResult(
                    results=[
                        (
                            {"msg": {"errors": [], "signed_claim": "signed:'claim'"}},
                            {"fake": "headers"},
                        )
                    ],
                    signing_keys=["test-signing-key"],
                    failed=False,
                ),
            )


@patch("uuid.uuid4", return_value="1234-5678-abcd-efgh")
def test_clear_sign_recv_timeout(patched_uuid, f_config_msg_signer_ok):
    clear_sign_operation = ClearSignOperation(
        inputs=["hello world"], signing_keys=["test-signing-key"], task_id="1", repo="repo"
    )

    with patch("pubtools.sign.signers.msgsigner.SendClient") as patched_send_client:
        with patch("pubtools.sign.signers.msgsigner.RecvClient") as patched_recv_client:
            patched_send_client.return_value.run.return_value = []
            errors = [
                MsgError(
                    name="MessagingTimeout",
                    description="Out of time when receiving messages",
                    source="test-source",
                ),
                MsgError(
                    name="MessagingTimeout",
                    description="Out of time when receiving messages",
                    source="test-source",
                ),
                MsgError(
                    name="MessagingTimeout",
                    description="Out of time when receiving messages",
                    source="test-source",
                ),
            ]
            patched_recv_client.return_value.get_errors.return_value = errors
            patched_recv_client.return_value.recv = {"1234-5678-abcd-efgh": "signed:'hello world'"}

            signer = MsgSigner()
            signer.load_config(load_config(f_config_msg_signer_ok))
            signer.retries = 2
            signer.send_retries = 1

            res = signer.clear_sign(clear_sign_operation)

            assert res == SigningResults(
                signer=signer,
                operation=clear_sign_operation,
                signer_results=MsgSignerResults(
                    status="error",
                    error_message="MessagingTimeout : Out of time when receiving messages\n"
                    "MessagingTimeout : Out of time when receiving messages\n",
                ),
                operation_result=ClearSignResult(outputs=[""], signing_keys=["test-signing-key"]),
            )


def test_recv_client_recv_message_break(
    f_cleanup_msgsigner_messages,
    f_qpid_broker,
    f_msgsigner_listen_to_topic,
    f_fake_msgsigner,
    f_msgsigner_send_to_queue,
    f_client_certificate,
    f_ca_certificate,
    f_config_msg_signer_ok2,
):
    qpid_broker, port = f_qpid_broker
    container_sign_operation = ContainerSignOperation(
        digests=("some-digest",),
        references=("some/reference:some-tag",),
        signing_keys=["test-signing-key"],
        task_id="1",
    )
    with patch(
        "pubtools.sign.clients.msg_recv_client.RecvClient.get_errors", autospec=True
    ) as patched_recv_get_errors, patch(
        "pubtools.sign.clients.msg_recv_client.RecvClient.get_received", autospec=True
    ) as patched_recv_get_received, patch(
        "uuid.uuid4", return_value="1234-5678-abcd-efgh"
    ) as _:
        patched_recv_get_errors.side_effect = [
            [
                MsgError(
                    name="MessagingTimeout",
                    description="Out of time when receiving messages",
                    source=ANY,
                )
            ],
            [
                MsgError(
                    name="MessagingTimeout",
                    description="Out of time when receiving messages",
                    source=ANY,
                )
            ],
            [
                MsgError(
                    name="MessagingTimeout",
                    description="Out of time when receiving messages",
                    source=ANY,
                )
            ],
            [
                MsgError(
                    name="MessagingTimeout",
                    description="Out of time when receiving messages",
                    source=ANY,
                )
            ],
            [],
        ]
        patched_recv_get_received.side_effect = [{"1234-5678-abcd-efgh": True}]
        signer = MsgSigner()
        signer.load_config(load_config(f_config_msg_signer_ok2))
        signer.retries = 2
        signer.send_retries = 1
        res = signer.container_sign(container_sign_operation)

        assert res == SigningResults(
            signer=signer,
            operation=container_sign_operation,
            signer_results=MsgSignerResults(
                status="error",
                error_message="MessagingTimeout : Out of time when receiving messages\n",
            ),
            operation_result=ContainerSignResult(
                results=[""], signing_keys=["test-signing-key"], failed=False
            ),
        )


def test_msgsig_doc_arguments():
    assert MsgSigner.doc_arguments() == {
        "options": {
            "messaging_brokers": {"description": "List of brokers URLS"},
            "messaging_cert_key": {
                "description": "Client certificate + key for messaging authorization"
            },
            "messaging_ca_cert": {"description": "Messaging CA certificate"},
            "topic_send_to": {"description": "Topic where to send the messages"},
            "topic_listen_to": {"description": "Topic where to listen for replies"},
            "creator": {"description": "Identification of creator of signing request"},
            "environment": {"description": "Environment indetification in sent messages"},
            "service": {"description": "Service identificator"},
            "task_id_attribute": {
                "description": "Attribute used to custom identification of signing request"
            },
            "timeout": {"description": "Timeout for messaging receive"},
            "retries": {"description": "Retries for messaging receive"},
            "send_retries": {"description": "Retries for messaging send+receive"},
            "message_id_key": {
                "description": "Attribute name in message body which should be used as message id"
            },
            "log_level": {"description": "Log level"},
            "key_aliases": {"description": "Aliases for signing keys"},
        },
        "examples": {
            "msg_signer": {
                "messaging_brokers": ["amqps://broker-01:5671", "amqps://broker-02:5671"],
                "messaging_cert_key": "~/messaging/cert.pem",
                "messaging_ca_cert": "~/messaging/ca_cert.crt",
                "topic_send_to": "topic://Topic.sign",
                "topic_listen_to": "queue://Consumer.{{creator}}.{{task_id}}.Topic.sign."
                "{{task_id}}",
                "creator": "pubtools-sign",
                "environment": "prod",
                "service": "pubtools-sign",
                "task_id_attribute": "task_id",
                "timeout": 1,
                "retries": 3,
                "send_retries": 2,
                "message_id_key": "123",
                "log_level": "debug",
                "key_aliases": "{'production':'abcde1245'}",
            }
        },
    }


def test_msgsigresult_to_dict():
    assert MsgSignerResults(status="status", error_message="error_message").to_dict() == {
        "status": "status",
        "error_message": "error_message",
    }


def test_msgsigresult_doc_arguments():
    assert MsgSignerResults.doc_arguments() == {
        "signer_result": {
            "type": "dict",
            "description": "Signing result status.",
            "returned": "always",
            "sample": {"status": "ok", "error_message": ""},
        }
    }


@patch("uuid.uuid4", side_effect=["1234-5678-abcd-efgh", "2234-5678-abcd-efgh"])
def test_blob_sign(patched_uuid, f_config_msg_signer_ok, f_client_certificate, f_ca_certificate):
    blob_sign_operation = BlobSignOperation(
        task_id="1",
        blobs=[b"blob1", b"blob2"],
        signing_keys=["test-signing-key"],
    )

    with patch("pubtools.sign.signers.msgsigner.SendClient") as patched_send_client:
        with patch("pubtools.sign.signers.msgsigner.RecvClient") as patched_recv_client:
            patched_send_client.return_value.run.return_value = []
            patched_recv_client.return_value.run.return_value = []
            patched_recv_client.return_value.recv = {
                "1234-5678-abcd-efgh": (
                    {"msg": {"errors": [], "signed_payload": "signed:'blob1'"}},
                    {"fake": "headers"},
                ),
                "2234-5678-abcd-efgh": (
                    {"msg": {"errors": [], "signed_payload": "signed:'blob2'"}},
                    {"fake": "headers"},
                ),
            }
            patched_recv_client.return_value.get_errors.return_value = []

            signer = MsgSigner()
            signer.load_config(load_config(f_config_msg_signer_ok))
            res = signer.blob_sign(blob_sign_operation)

            patched_send_client.assert_called_with(
                messages=[ANY, ANY],
                broker_urls=["amqps://broker-01:5671", "amqps://broker-02:5671"],
                cert=f_client_certificate,
                ca_cert=f_ca_certificate,
                retries=3,
                errors=patched_recv_client.return_value.get_errors.return_value,
            )

            assert res == SigningResults(
                signer=signer,
                operation=blob_sign_operation,
                signer_results=MsgSignerResults(status="ok", error_message=""),
                operation_result=BlobSignResult(
                    results=[
                        (
                            {"msg": {"errors": [], "signed_payload": "signed:'blob1'"}},
                            {"fake": "headers"},
                        ),
                        (
                            {"msg": {"errors": [], "signed_payload": "signed:'blob2'"}},
                            {"fake": "headers"},
                        ),
                    ],
                    signing_keys=["test-signing-key"],
                    failed=False,
                ),
            )


@patch("uuid.uuid4", side_effect=["1234-5678-abcd-efgh", "2234-5678-abcd-efgh"])
def test_blob_sign_multiple_signing_keys(
    patched_uuid, f_config_msg_signer_ok, f_client_certificate, f_ca_certificate
):
    blob_sign_operation = BlobSignOperation(
        task_id="1",
        blobs=[b"blob1"],
        signing_keys=["test-signing-key", "test-signing-key-1"],
    )

    with patch("pubtools.sign.signers.msgsigner.SendClient") as patched_send_client:
        with patch("pubtools.sign.signers.msgsigner.RecvClient") as patched_recv_client:
            patched_send_client.return_value.run.return_value = []
            patched_recv_client.return_value.run.return_value = []
            patched_recv_client.return_value.recv = {
                "1234-5678-abcd-efgh": (
                    {"msg": {"errors": [], "signed_payload": "signed:'blob1'"}},
                    {"fake": "headers"},
                ),
                "2234-5678-abcd-efgh": (
                    {"msg": {"errors": [], "signed_payload": "signed:'blob1'"}},
                    {"fake": "headers"},
                ),
            }
            patched_recv_client.return_value.get_errors.return_value = []

            signer = MsgSigner()
            signer.load_config(load_config(f_config_msg_signer_ok))
            res = signer.blob_sign(blob_sign_operation)

            patched_send_client.assert_called_with(
                messages=[ANY, ANY],
                broker_urls=["amqps://broker-01:5671", "amqps://broker-02:5671"],
                cert=f_client_certificate,
                ca_cert=f_ca_certificate,
                retries=3,
                errors=patched_recv_client.return_value.get_errors.return_value,
            )

            assert res == SigningResults(
                signer=signer,
                operation=blob_sign_operation,
                signer_results=MsgSignerResults(status="ok", error_message=""),
                operation_result=BlobSignResult(
                    results=[
                        (
                            {"msg": {"errors": [], "signed_payload": "signed:'blob1'"}},
                            {"fake": "headers"},
                        ),
                        (
                            {"msg": {"errors": [], "signed_payload": "signed:'blob1'"}},
                            {"fake": "headers"},
                        ),
                    ],
                    signing_keys=["test-signing-key", "test-signing-key-1"],
                    failed=False,
                ),
            )


def test_msg_blob_sign_error(f_msg_signer, f_config_msg_signer_ok):
    f_msg_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "error",
        "error_message": "simulated error",
    }
    f_msg_signer.return_value.sign.return_value.operation_result.results = []
    f_msg_signer.return_value.sign.return_value.operation_result.signing_keys = []
    f_msg_signer.return_value.sign.return_value.operation.to_dict.return_value = {}
    with tempfile.NamedTemporaryFile("w") as blob_file:
        result = CliRunner().invoke(
            msg_blob_sign_main,
            [
                "--signing-key",
                "test-signing-key",
                "--blob-file",
                blob_file.name,
                "--task-id",
                "1",
                "--config-file",
                f_config_msg_signer_ok,
            ],
        )
    print(result.stdout)
    assert result.exit_code == 1, result.output


def test_msg_blob_sign_error_raw(f_msg_signer, f_config_msg_signer_ok):
    f_msg_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "error",
        "error_message": "simulated error",
    }
    f_msg_signer.return_value.sign.return_value.operation_result.results = [
        ({"i": 456, "msg": {"errors": ["error1"], "signed_payload": "signed"}}, {})
    ]
    f_msg_signer.return_value.sign.return_value.operation_result.signing_keys = []
    f_msg_signer.return_value.sign.return_value.operation.to_dict.return_value = {}
    with tempfile.NamedTemporaryFile("w") as blob_file:
        result = CliRunner().invoke(
            msg_blob_sign_main,
            [
                "--raw",
                "--signing-key",
                "test-signing-key",
                "--blob-file",
                blob_file.name,
                "--task-id",
                "1",
                "--config-file",
                f_config_msg_signer_ok,
            ],
        )
    print(result.stdout)
    assert result.exit_code == 1, result.output


@patch("uuid.uuid4", return_value="1234-5678-abcd-efgh")
def test_blob_sign_recv_errors(patched_uuid, f_config_msg_signer_ok):
    blob_sign_operation = BlobSignOperation(
        task_id="1",
        blobs=[b"blob1"],
        signing_keys=["test-signing-key", "test-signing-key-1"],
    )

    with patch("pubtools.sign.signers.msgsigner.SendClient") as patched_send_client:
        with patch("pubtools.sign.signers.msgsigner.RecvClient") as patched_recv_client:
            patched_send_client.return_value.run.return_value = []
            errors = [
                MsgError(
                    name="TestError", description="test error description", source="test-source"
                )
            ]
            patched_recv_client.return_value._errors = errors
            patched_recv_client.return_value.get_errors.return_value = errors
            patched_recv_client.return_value.recv = {"1234-5678-abcd-efgh": "signed:'hello world'"}

            signer = MsgSigner()
            signer.load_config(load_config(f_config_msg_signer_ok))
            res = signer.sign(blob_sign_operation)

            assert res == SigningResults(
                signer=signer,
                operation=blob_sign_operation,
                signer_results=MsgSignerResults(
                    status="error", error_message="TestError : test error description\n"
                ),
                operation_result=BlobSignResult(
                    results=["", ""],
                    signing_keys=["test-signing-key", "test-signing-key-1"],
                    failed=False,
                ),
            )


def test_msg_blob_sign_batch(f_msg_signer, f_config_msg_signer_ok):
    f_msg_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {}
    f_msg_signer.return_value.sign.return_value.operation_result.results = []
    f_msg_signer.return_value.sign.return_value.operation_result.signing_keys = []
    f_msg_signer.return_value.sign.return_value.operation.to_dict.return_value = {}
    with tempfile.NamedTemporaryFile("w") as blob_file:
        result = CliRunner().invoke(
            msg_blob_sign_main,
            [
                "--raw",
                "--signing-key",
                "test-signing-key",
                "--blob-file",
                blob_file.name,
                "--task-id",
                "1",
                "--signer-type",
                "batch",
                "--config-file",
                f_config_msg_signer_ok,
            ],
        )
    print(result.stdout)
    assert result.exit_code == 1, result.output


def test_msg_blob_sign_raw(f_msg_signer, f_config_msg_signer_ok):
    f_msg_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "ok"
    }
    f_msg_signer.return_value.sign.return_value.operation_result.results = [
        ({"i": 456, "msg": {"errors": [], "signed_payload": "signed"}}, {})
    ]
    f_msg_signer.return_value.sign.return_value.operation_result.signing_keys = []
    f_msg_signer.return_value.sign.return_value.operation.to_dict.return_value = {}
    with tempfile.NamedTemporaryFile("w") as blob_file:
        blob_file.write("blob1\nblob2\n")
        blob_file.flush()
        result = CliRunner().invoke(
            msg_blob_sign_main,
            [
                "--raw",
                "--signing-key",
                "test-signing-key",
                "--blob-file",
                blob_file.name,
                "--task-id",
                "1",
                "--config-file",
                f_config_msg_signer_ok,
            ],
        )
    print(result.stdout)
    assert result.exit_code == 0, result.output


def test_msg_blob_sign_raw_error_messages(f_msg_signer, f_config_msg_signer_ok):
    f_msg_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "ok"
    }
    f_msg_signer.return_value.sign.return_value.operation_result.results = [
        ({"i": 456, "msg": {"errors": ["error1"], "signed_claim": "signed"}}, {})
    ]
    f_msg_signer.return_value.sign.return_value.operation_result.signing_keys = []
    f_msg_signer.return_value.sign.return_value.operation.to_dict.return_value = {}
    with tempfile.NamedTemporaryFile("w") as blob_file:
        blob_file.write("blob1\nblob2\n")
        blob_file.flush()
        result = CliRunner().invoke(
            msg_blob_sign_main,
            [
                "--raw",
                "--signing-key",
                "test-signing-key",
                "--blob-file",
                blob_file.name,
                "--task-id",
                "1",
                "--config-file",
                f_config_msg_signer_ok,
            ],
        )
    print(result.stdout)
    assert result.exit_code == 1, result.output


def test_msg_blob_sign_requester(f_msg_signer, f_config_msg_signer_ok):
    f_msg_signer.return_value.sign.return_value.signer_results.to_dict.return_value = {
        "status": "ok"
    }
    f_msg_signer.return_value.sign.return_value.operation_result.results = []
    f_msg_signer.return_value.sign.return_value.operation_result.signing_keys = []
    f_msg_signer.return_value.sign.return_value.operation.to_dict.return_value = {}
    with tempfile.NamedTemporaryFile("w") as blob_file:
        blob_file.write("blob1\nblob2\n")
        blob_file.flush()
        result = CliRunner().invoke(
            msg_blob_sign_main,
            [
                "--raw",
                "--signing-key",
                "test-signing-key",
                "--blob-file",
                blob_file.name,
                "--task-id",
                "1",
                "--config-file",
                f_config_msg_signer_ok,
                "--requester",
                "test-requester",
            ],
        )
    print(result.stdout)
    assert result.exit_code == 0, result.output
